import sys
import re
import argparse
from typing import Optional, List
import os
import re
import time
import sys
from typing import List, Dict, Optional
from sqlalchemy import create_engine, inspect, MetaData, Table
from sqlalchemy.engine import reflection
from collections import namedtuple


class BaseHelper(object):
    @classmethod
    def run(cls, commands: list):
        raise NotImplementedError("Need Implement")


def generate_hive_create(self, args):
    pass


def generate_partition_name(pt_name):
    return f"""PARTITIONED BY (
         {pt_name} STRING COMMENT '分区字段'
    )"""


def create_hive_table(name: str, columns: List[dict], table_comment: Optional[str],
                      output, hive_database, partition, pt_name):
    print(f'use {hive_database};', file=output)
    print(f'CREATE TABLE IF NOT EXISTS `{name}` (', file=output)
    is_first = True
    for col in columns:
        print(('' if is_first else ',\n') + " " * 4, end='', file=output)
        if is_first:
            is_first = False

        # print(col['type'], type(col['type']))
        _t = col['type']
        _n = col['name']
        _c = col['comment']
        t_name = str(_t)
        simple = re.findall(r"(\w+\(\d+,?\s*\d*\))\s*", t_name)
        if simple:
            r_name = simple[0]
        else:
            r_name = t_name.split(' ')[0]

        if r_name.startswith('BIGINT') or r_name.startswith('INTEGER'):
            r_name = r_name.split('(')[0]

        if r_name.startswith('SMALLINT'):
            r_name = 'INT'

        if t_name.endswith('TEXT') or t_name.startswith('VARBINARY'):
            r_name = "STRING"

        if t_name.startswith('DATETIME'):
            r_name = 'TIMESTAMP'

        comment = f" COMMENT '{_c}'" if _c else ''
        column = f'`{_n}` {r_name}{comment}'
        print(column, file=output, end='')
    print('\n)', file=output)
    if table_comment:
        print(f'COMMENT "{table_comment}"', file=output)
    if partition:
        print(generate_partition_name(pt_name), file=output)
    print(';', file=output)


def dump_db2hive(name: str, columns: List[dict],
                 output, hive_database, partition, pt_name):
    pass


class DBToHive(BaseHelper):
    @classmethod
    def run(cls, commands: list):
        parser = argparse.ArgumentParser(description='Generate daily backup data from database To Hive SQL')
        parser.add_argument('--database', help='database name')
        parser.add_argument('--host', help='database host')
        parser.add_argument('--user', help='database user')
        parser.add_argument('--password', help='database user password', default=None)
        parser.add_argument('--table', help='backup table name', default='*')
        parser.add_argument('--port', help='database port', type=int, default=3306)
        parser.add_argument('--output', help='output file name if null then print', default=None)
        parser.add_argument('--hive_database', help='dump hive database if null then the db database name',
                            default=None)
        parser.add_argument('--hive_catalog', help='hive catalog setting in your flink sql config',
                            default='myhive')
        parser.add_argument('--partition', help='add partition table', action='store_true', default=False)
        parser.add_argument('--pt_name', help='partition name', default='pt')
        parser.add_argument('--db_type', help='database type', default='mysql', choices=['mysql', 'postgresql'])

        args = parser.parse_args(commands)
        if args.password is None:
            password = input('your database base password:')
            args.__setattr__('password', password.strip())
        if args.output is None:
            args.__setattr__('output', sys.stdout)
        else:
            args.__setattr__('output', open(args.output, 'w'))

        if args.hive_database is None:
            args.__setattr__('hive_database', args.database)
        driver = 'pymysql' if args.db_type == 'mysql' else 'psycopg2'
        db_url = f"{args.db_type}+{driver}://{args.user}:{args.password}@{args.host}:{args.port}/{args.database}?charset=utf8"

        engine = create_engine(db_url)
        insp = inspect(engine)
        all_tables_names = insp.get_table_names()
        tables = all_tables_names if args.table == '*' else args.table.split(',')

        need_tables = set()

        for x in tables:
            if x not in all_tables_names:
                print('table {} not found in database'.format(x))

        for n in need_tables:
            t_comment = insp.get_table_comment(n)['text']
            cs = insp.get_columns(n)
            create_hive_table(n, columns=cs, table_comment=t_comment, output=args.output,
                              hive_database=args.hive_database, partition=args.partition, pt_name=args.pt_name)

        for n in need_tables:
            cs = insp.get_columns(n)
            dump_db2hive(n, columns=cs, output=args.output,
                         hive_database=args.hive_database, partition=args.partition, pt_name=args.pt_name)

        # print(args)


if __name__ == '__main__':
    DBToHive.run(['--database', 'data', '--host', 'localhost', '--user', 'root', '--partition', '--password', 'abc',
                  '--hive_database', 'example'])



from sqlalchemy import create_engine


def execute(sql, url, *args) -> list:
    engine = create_engine(url)
    with engine.connect() as con:
        rs = con.execute(sql, *args)
        return list(rs)



import time
import logging
import logzero
from tornado import ioloop
from logzero import setup_logger
from collections import defaultdict
from typing import Callable, Optional
# from fsqlfly.models import Transform, auto_close
from requests import Session
from datetime import datetime, date
from fsqlfly.workflow import run_transform
from fsqlfly.settings import FSQLFLY_DEBUG, FSQLFLY_MAIL_ENABLE
from fsqlfly.utils.job_manage import JobControlHandle
from fsqlfly.contrib.mail import MailHelper
from fsqlfly.db_helper import DBSession, DBDao


def get_log_file(file: str):
    return setup_logger(name='JobDaemon', logfile=file, maxBytes=int(1e6), backupCount=3,
                        level=logging.DEBUG if FSQLFLY_DEBUG else logging.INFO)


class FlinkJobDaemon:

    def __init__(self,
                 flink_host: str,
                 max_try: int,
                 login_file: str,
                 max_req_try: int = 5,
                 stop_handle: Optional[Callable] = None):
        self.logger = get_log_file(login_file)
        self.flink_host = flink_host
        self.max_try = max_try
        self.stop_handle = stop_handle
        self.max_req_try = max_req_try
        self.session = Session()
        self.run_times = defaultdict(lambda: defaultdict(int))
        self.started_jobs = set()

    def request(self, func: Callable, try_times: int = 0):
        try:
            return func()
        except Exception as err:
            if try_times > self.max_req_try:
                print('Meet Exception please check FSQLFLY_FINK_HOST is correct')
                raise err
            else:
                return self.request(func, try_times + 1)

    @classmethod
    def send_email(cls, title: str = '', content: str = ''):
        if FSQLFLY_MAIL_ENABLE:
            print(MailHelper.send(title, content))

    def run(self):
        self.logger.debug('Start Running Flink Job Damon {}'.format(str(datetime.now())[:19]))

        today = str(date.today())
        start_time = time.time()
        session = DBSession.get_session()
        job_names = DBDao.get_job_names(session=session)

        living_job = JobControlHandle.live_job_names
        for k, transform in job_names.items():
            if k not in living_job:
                if self.run_times[today][k] > self.max_try:
                    self.logger.error('job run too many times one day {}'.format(k))
                    self.send_email('job run too many times one day {}'.format(k))
                else:
                    self.run_times[today][k] += 1
                    self.logger.info('job {} begin run '.format(k))
                    is_ok, r = run_transform(transform)
                    if not is_ok:
                        self.send_email('job start fail {}'.format(k), r)
                        self.logger.error(r)
                    else:
                        if k in self.started_jobs:
                            self.send_email('try restart job {}, last fail'.format(k), r)
                        else:
                            self.started_jobs.add(k)
                    if self.stop_handle:
                        self.stop_handle(k)

        end_time = time.time()

        cost = end_time - start_time

        self.logger.debug(
            " ".join([str(datetime.now())[:19], ' damon cost ', '%.2f' % cost, ' second', ' will sleep ']))
        session.close()

    def get_periodic_callback(self, period) -> Callable:
        def _warp():
            ioloop.PeriodicCallback(self.run, period * 1000).start()

        return _warp



# -*- coding:utf-8 -*-
from fsqlfly.common import safe_authenticated
from fsqlfly.base_handle import BaseHandler
from fsqlfly.common import DBRes
from fsqlfly.connection_manager import ManagerHelper


class ManagerHandler(BaseHandler):
    @safe_authenticated
    def post(self, model: str, mode: str, pk: str):
        if ManagerHelper.is_support(model, mode):
            self.write_res(ManagerHelper.run(model, mode, pk))
        else:
            self.write_res(DBRes.api_error("{} not support now".format(mode)))


default_handlers = [
    (r'/api/(?P<model>\w+)/(?P<mode>update|clean|init|list)/(?P<pk>[a-zA-Z_0-9]+)', ManagerHandler),
]



# -*- coding:utf-8 -*-
import sys
import argparse
import logzero
from os.path import join, dirname, abspath


def run_webserver(commands: list):
    extend_command = None
    if '--jobdaemon' in commands:
        from fsqlfly.job_manager.daemon import FlinkJobDaemon
        from fsqlfly import settings
        daemon = FlinkJobDaemon(settings.FSQLFLY_FINK_HOST,
                                settings.FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY,
                                settings.FSQLFLY_JOB_LOG_FILE)
        extend_command = daemon.get_periodic_callback(settings.FSQLFLY_JOB_DAEMON_FREQUENCY)
        logzero.logger.debug('add job daemon command {}: {}: {}: {}'.format(settings.FSQLFLY_FINK_HOST,
                                                                            settings.FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY,
                                                                            settings.FSQLFLY_JOB_LOG_FILE,
                                                                            settings.FSQLFLY_JOB_DAEMON_FREQUENCY))

    from fsqlfly.app import run_web
    try:
        run_web(extend_command=extend_command)
    except KeyboardInterrupt:
        logzero.logger.info("Stop Web...")


def init_db(commands: list):
    from fsqlfly.db_helper import DBDao

    DBDao.create_all_tables()


def reset_db(commands: list):
    from fsqlfly.db_helper import DBDao
    conformed_parser = argparse.ArgumentParser("Conformed")
    conformed_parser.add_argument('-f', '--force', type=bool, default=False, help='force running')
    args = conformed_parser.parse_args(commands)
    DBDao.delete_all_tables(force=args.force)


def run_echo_env(commands: list):
    if len(commands) > 0:
        out = open(commands[0], 'w')
    else:
        out = sys.stdout

    lib_path = dirname(abspath(__file__))
    print(open(join(lib_path, 'env.template'), 'r').read(), file=out)


def run_db2hive(comands: list):
    from fsqlfly.helper import DBToHive
    DBToHive.run(comands)


def run_canal(commands: list):
    from fsqlfly.contrib.canal import Consumer
    Consumer.build(commands[0]).run()


def main():
    support_command = {
        "echoenv": run_echo_env,
        "webserver": run_webserver,
        "initdb": init_db,
        "resetdb": reset_db,
        "runcanal": run_canal
    }
    args = sys.argv[1:]
    method = args[0] if len(args) > 0 else 'help'
    if len(args) == 0 or method in ('-h', 'help') or method not in support_command:
        print("Usage : fsqlfly [-h] {}".format('|'.join(list(support_command.keys()) + ['help'])))
        return
    process = support_command[args[0]]
    process(sys.argv[2:])


if __name__ == '__main__':
    main()



import json
import logging
from copy import deepcopy
from datetime import datetime, date
from airflow.operators.sensors import BaseSensorOperator
from airflow.hooks.http_hook import HttpHook
from airflow.utils.decorators import apply_defaults


class _BaseJobOperator(BaseSensorOperator):
    template_fields = ['data', 'headers']
    RUN_STATUS = 'RUNNING'
    FINISHED_STATUS = 'FINISHED'

    @classmethod
    def gen_job_url(cls, job_name, method):
        return f'/api/transform/{method}/{job_name}'

    def get_connector_job_list(self, job_name):
        url = f'/api/connector/list/{job_name}'
        return self.http.run(url, headers=self.headers).json()['data']

    def get_all_job_names(self):
        raise NotImplementedError

    def get_job_list(self):
        if self.all_jobs is None:
            self.all_jobs = self.get_all_job_names()
        left_jobs = [x for x in self.all_jobs if x not in self.finished_jobs and x not in self.job_pools]
        return left_jobs

    def get_start_endpoint(self, job_name):
        return self.gen_job_url(job_name, 'start')

    def get_status_endpoint(self, job_name):
        return self.gen_job_url(job_name, 'status')

    def complete_job(self, job_name):
        self.finished_jobs.append(job_name)

    @apply_defaults
    def __init__(self, http_conn_id, token, job_name,
                 data=None, headers=None, method='start', daemon=True, parallelism=0, *args, **kwargs):
        basic_headers = {'Content-Type': "application/json",
                         'Token': token}
        if headers:
            basic_headers.update(headers)
        self.headers = basic_headers
        self.http_conn_id = http_conn_id
        self.job_name = job_name
        self.http = HttpHook('POST', http_conn_id=self.http_conn_id)

        self.data = data if data is not None else {}
        self.job_last_run_id = dict()
        self.job_pools = []
        self.all_jobs = None
        self.finished_jobs = []
        self.parallelism = parallelism
        self.method = method
        self.daemon = daemon

        super(_BaseJobOperator, self).__init__(*args, **kwargs)

    def get_req_data(self, job_name):
        def _parse_date_time(d):
            if isinstance(d, datetime) or isinstance(d, date):
                return str(d)

        send_data = deepcopy(self.data)
        last_run_job_id = self.job_last_run_id.get(job_name)
        if last_run_job_id:
            send_data['last_run_job_id'] = last_run_job_id
        return json.dumps(send_data, ensure_ascii=True, default=_parse_date_time)

    def run_other_mode(self):
        for job_name in self.get_job_list():
            res = self.http.run(self.gen_job_url(job_name, self.method)
                                , data=self.data, headers=self.headers).json()
            if not res['success']:
                raise Exception('{} Job Fail response: {}'.format(self.method, str(res)))
            else:
                print('Job {} {} Finished'.format(self.job_name, self.method))

    def execute(self, context):
        if self.method != 'start':
            return self.run_other_mode()
        super(_BaseJobOperator, self).execute(context)

    def get_job_status(self, job_name):
        res = self.http.run(self.get_status_endpoint(job_name), data=self.get_req_data(job_name),
                            headers=self.headers).json()
        full_msg = "req: {} code: {} msg: {}".format(job_name, res['code'], res['msg'])
        if not res['success']:
            raise Exception(full_msg)
        msg = res['msg']
        if msg.endswith(self.RUN_STATUS):
            job_id, _ = msg.split('_' + self.RUN_STATUS, 1)
            self.job_last_run_id[job_name] = job_id
        return msg

    def add_job_to_pool(self, job_name):
        status = self.get_job_status(job_name)
        if status.endswith(self.RUN_STATUS):
            raise Exception("Job {} Already {}".format(job_name, status))
        self.job_pools.append(job_name)

        res = self.http.run(self.get_start_endpoint(job_name), data=self.get_req_data(job_name),
                            headers=self.headers).json()
        if not res['success']:
            raise Exception('Start Job Fail response: {}'.format(str(res)))

    def pop_finished_job(self):
        current_size = len(self.job_pools)
        for _ in range(current_size):
            job_name = self.job_pools.pop(0)
            if self.daemon:
                msg = self.get_job_status(job_name)
                if msg == self.FINISHED_STATUS:
                    self.finished_jobs.append(job_name)
                elif msg.endswith(self.RUN_STATUS):
                    logging.debug("Wait For :" + msg)
                    self.job_pools.append(job_name)
                else:
                    err_info = "Job {} Fail With Other Exception: {}".format(job_name, msg)
                    logging.error(err_info)
                    raise Exception(err_info)
            else:
                self.finished_jobs.append(job_name)

    def job_pool_finished(self):
        self.pop_finished_job()
        if self.get_job_list() or self.job_pools:
            return False
        return True

    def poke(self, context):
        run_jobs = self.get_job_list()
        while (len(self.job_pools) < self.parallelism or self.parallelism == 0) and run_jobs:
            job_name = run_jobs.pop()
            self.add_job_to_pool(job_name)
        return self.job_pool_finished()


class FSQLFlyOperator(_BaseJobOperator):
    def get_all_job_names(self):
        return [self.job_name]


class FSQLFlyConnectorOperator(_BaseJobOperator):
    def get_all_job_names(self):
        return self.get_connector_job_list(self.job_name)



# -*- coding:utf-8 -*-
import time
from io import StringIO
from typing import Dict, Optional, Any
from fsqlfly.common import safe_authenticated
from collections import defaultdict
from fsqlfly.settings import FSQLFLY_FINK_HOST, TEMP_TERMINAL_HEAD
from fsqlfly.base_handle import BaseHandler
from fsqlfly.utils.job_manage import JobControlHandle, handle_job
from fsqlfly.db_helper import DBDao
from fsqlfly.common import DBRes


class JobHandler(BaseHandler):
    @safe_authenticated
    def get(self, mode: str, pk: str):
        return self.write_res(handle_job(mode, pk, self.json_body))

    post = get


def get_latest_transform(refresh_seconds: int = 5) -> Dict[str, dict]:
    def _get():
        job_infos = defaultdict(dict)
        for x in DBDao.get_transform():
            name = "{}_{}".format(x.id, x.name)
            job_infos[name] = x.to_dict()
        return job_infos

    __name = '__transforms'
    if not hasattr(get_latest_transform, __name):
        setattr(get_latest_transform, __name, (_get(), time.time()))
    objects, t = getattr(get_latest_transform, __name)
    if time.time() - t > refresh_seconds:
        objects = _get()
        setattr(get_latest_transform, __name, (objects, time.time()))
    return objects


class JobList(BaseHandler):
    @safe_authenticated
    def get(self):
        job_infos = get_latest_transform()
        all_jobs = list()
        for job in JobControlHandle.job_status:
            if not JobControlHandle.is_real_job(job.name):
                continue
            base = dict(**job._asdict())
            if job.name in job_infos:
                base.update(job_infos[job.name])
                base['t_id'] = job_infos[job.name]['id']
            elif job.name.startswith(TEMP_TERMINAL_HEAD):
                base['name'] = 'TEMPORARY'
                base['url'] = '/terminal/{}'.format(job.name[len(TEMP_TERMINAL_HEAD):])

            base['id'] = job.job_id
            base['detail_url'] = FSQLFLY_FINK_HOST + '/#/job/{}/overview'.format(job.job_id)

            all_jobs.append(base)

        return self.write_res(DBRes(data=all_jobs))


default_handlers = [
    (r'/api/job/(?P<mode>\w+)/(?P<pk>[a-zA-Z0-9]+)', JobHandler),
    (r'/api/job', JobList),
]



# -*- coding:utf-8 -*-
import re
import functools
import urllib
import traceback
import attr
from urllib.parse import urlencode
from re import _pattern_type
from typing import Any, Optional, NamedTuple, Callable, Awaitable, Union, List, Type
from tornado.web import RequestHandler, HTTPError
from logzero import logger

SUPPORT_TABLE_TYPE = {'sink', 'source', 'both', 'view', 'temporal-table'}
SUPPORT_MANAGER = {'hive', 'jdbc', 'kafka', 'hbase', 'elasticsearch', 'filesystem'}
CONNECTOR_TYPE = {'canal', 'system'}


class CodeMsg(NamedTuple):
    code: int
    msg: str


class RespCode:
    Success = CodeMsg(200, 'Success')
    ServerError = CodeMsg(500, 'Web Server Error')
    NeedLogin = CodeMsg(503, 'You Need Login')
    LoginFail = CodeMsg(501, 'Wrong Password or Token')
    APIFail = CodeMsg(502, 'Invalid API Request')
    InvalidHttpMethod = CodeMsg(405, 'Invalid HTTP method.')


@attr.s
class DBRes:
    data: Any = attr.ib(default=None)
    code: int = attr.ib(default=200)
    msg: Optional[str] = attr.ib(default=None)
    success: bool = attr.ib()

    @success.default
    def get_success(self):
        return self.code == 200

    @classmethod
    def login_error(cls, msg: str = RespCode.LoginFail.code):
        return DBRes(code=RespCode.LoginFail.code, msg=msg)

    @classmethod
    def api_error(cls, msg: str = RespCode.APIFail.msg):
        return DBRes(code=RespCode.APIFail.code, msg=msg)

    @classmethod
    def not_found(cls, msg: str = "Can't found Object"):
        return DBRes(code=RespCode.APIFail.code, msg=msg)

    @classmethod
    def resource_locked(cls, msg: str = "Resource Locked"):
        return DBRes(code=RespCode.APIFail.code, msg=msg)

    @classmethod
    def sever_error(cls, msg: str = RespCode.ServerError.msg):
        return DBRes(code=RespCode.ServerError.code, msg=msg)


def safe_authenticated(
        method: Callable[..., Optional[Awaitable[None]]]
) -> Callable[..., Union[Optional[Awaitable[None]], DBRes]]:
    """Decorate methods with this to require that the user be logged in.

    If the user is not logged in, they will be redirected to the configured
    `login url <RequestHandler.get_login_url>`.

    If you configure a login url with a query parameter, Tornado will
    assume you know what you're doing and use it as-is.  If not, it
    will add a `next` parameter so the login page knows where to send
    you once you're logged in.
    """

    @functools.wraps(method)
    def wrapper(  # type: ignore
            self: RequestHandler, *args, **kwargs
    ) -> Union[Optional[Awaitable[None]], DBRes]:
        if not self.current_user:
            if self.request.method in ("GET", "HEAD"):
                url = self.get_login_url()
                if "?" not in url:
                    if urllib.parse.urlsplit(url).scheme:
                        # if login url is absolute, make next absolute too
                        next_url = self.request.full_url()
                    else:
                        assert self.request.uri is not None
                        next_url = self.request.uri
                    url += "?" + urlencode(dict(next=next_url))
                self.redirect(url)
                return None
            raise HTTPError(403)
        from fsqlfly import settings
        if settings.FSQLFLY_DEBUG:
            return method(self, *args, **kwargs)
        try:
            return method(self, *args, **kwargs)
        except Exception:
            err = traceback.format_exc()
            logger.error(err)
            return DBRes.sever_error(msg=f'meet {err}')

    return wrapper


class NameFilter:
    @classmethod
    def get_pattern(cls, s: str) -> List[Type[_pattern_type]]:
        s = s.replace('，', ',')
        patterns = [s.strip()] if ',' not in s else s.split(',')
        res = [re.compile(pattern=x + '$') for x in patterns]
        return res

    def __init__(self, include: str = '', exclude: str = ''):
        self.includes = self.get_pattern(include) if include else [re.compile(r'.*')]
        self.excludes = self.get_pattern(exclude) if exclude else []

    def __contains__(self, item: str) -> bool:
        if any(map(lambda x: x.match(item) is not None, self.includes)):
            if self.excludes and any(map(lambda x: x.match(item), self.excludes)):
                return False
            return True
        return False


@attr.s
class SchemaField:
    name: str = attr.ib()
    type: str = attr.ib()
    comment: Optional[str] = attr.ib(default=None)
    nullable: Optional[bool] = attr.ib(default=True)
    autoincrement: Optional[bool] = attr.ib(default=None)
    rowtime: Optional[Any] = attr.ib(default=None)
    proctime: Optional[bool] = attr.ib(default=None)


@attr.s
class SchemaContent:
    name: str = attr.ib()
    type: str = attr.ib()
    database: Optional[str] = attr.ib(default=None)
    comment: Optional[str] = attr.ib(default=None)
    primary_key: Optional[str] = attr.ib(default=None)
    fields: List[SchemaField] = attr.ib(factory=list)
    partitionable: bool = attr.ib(default=False)


@attr.s
class VersionConfig:
    exclude: Optional[str] = attr.ib(default=None)
    include: Optional[str] = attr.ib(default=None)
    update_mode: Optional[str] = attr.ib(default=None)
    query: Optional[str] = attr.ib(default=None)
    history_table: Optional[str] = attr.ib(default=None)
    primary_key: Optional[str] = attr.ib(default=None)
    time_attribute: Optional[str] = attr.ib(default=None)
    format: Optional[Any] = attr.ib(default=None)
    schema: list = attr.ib(factory=list)


class BlinkSQLType:
    STRING = 'STRING'
    BOOLEAN = 'BOOLEAN'
    BYTES = 'BYTES'
    DECIMAL = 'DECIMAL(38,18)'
    TINYINT = 'TINYINT'
    SMALLINT = 'SMALLINT'
    INTEGER = 'INTEGER'
    BIGINT = 'BIGINT'
    FLOAT = 'FLOAT'
    DOUBLE = 'DOUBLE'
    DATE = 'DATE'
    TIME = 'TIME'
    TIMESTAMP = 'TIMESTAMP(3)'


class BlinkHiveSQLType(BlinkSQLType):
    INTERVAL = 'INTERVAL'
    ARRAY = 'ARRAY'
    MULTISET = 'MULTISET'
    MAP = 'MAP'
    RAW = 'RAW'


class BlinkTableType:
    sink = 'sink'
    source = 'source'
    both = 'both'


class DBSupportType:
    MySQL = 'MySQL'
    PostgreSQL = 'PostgreSQL'


class CanalMode:
    upsert = 'upsert'
    update = 'update'
    insert = 'insert'
    delete = 'delete'
    all = 'all'

    @classmethod
    def support_modes(cls) -> set:
        return {cls.upsert, cls.insert, cls.update, cls.delete}

    def __contains__(self, item) -> bool:
        return item in self.support_modes

    def __init__(self, config_mode: str):
        if CanalMode.all in config_mode:
            mode = [CanalMode.update, CanalMode.insert, CanalMode.delete]
        else:
            mode = config_mode.split(',')
        all_mode_support = all(map(lambda x: x in CANAL_MODE, mode))
        assert all_mode_support, 'Canal Config not support mode {}'.format(mode)
        if CanalMode.upsert in mode:
            assert len(mode) == 1, 'upsert only support run by itself'
        self._mode = mode

    def is_upsert(self):
        return self.upsert in self._mode

    def is_support(self, mode: str) -> bool:
        if self.is_upsert():
            return True
        return mode in self._mode

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        if self._i >= len(self._mode):
            raise StopIteration
        v = self._mode[self._i]
        self._i += 1
        return v


CANAL_MODE = CanalMode.support_modes()

DEFAULT_CONFIG = """
[jdbc]
insert_primary_key = false
add_read_partition_key = false
auto_partition_bound = false
read_partition_fetch_size = 100
read_partition_lower_bound = 0
read_partition_upper_bound = 50000
read_partition_num = 50

[kafka]
startup_mode = latest-offset

[hive]

example = 1

[canal]
mode = upsert
process_time_enable = true
process_time_name = flink_process_time
rowtime_enable = true
rowtime_name = mysql_row_time
rowtime_watermarks = 5000
rowtime_from = MYSQL_DB_EXECUTE_TIME
binlog_type_name = MYSQL_DB_EVENT_TYPE
before_column_suffix = _before
after_column_suffix = _after
update_suffix = _updated
table_filter = .*\..*

[system]
execution_parallelism =  1
source_include =  .*
source_exclude =  ''
target_database_format =  {{ resource_name.database }}
target_table_format =  {{ resource_name.name }}
transform_name_format =  {{ source_type }}2{{ target_type }}__{{ connector.name }}__{{ resource_name.database }}__{{ resource_name.name }}  
use_partition =  false
partition_name =  pt
partition_value =  {{ ds_nodash }}
overwrite =  false

"""



from fsqlfly.common import *
import time
import os
import logging
import sys
import json
from logzero import logger
from typing import Optional, Dict, Tuple
from collections import defaultdict
from datetime import datetime
from random import randint
from configparser import ConfigParser, SectionProxy
from argparse import ArgumentParser
from fsqlfly.db_helper import DBSession, Connector, Connection, Session, ResourceVersion

PRINT_EACH = 100
WAIT_TIMES = 0.5


class Consumer:
    def __init__(self, canal_mode: CanalMode, bootstrap_servers: str, topics: Dict[Tuple[str, str, str], str], *args,
                 table_filter: Optional[str], canal_host: str, canal_port: str, canal_username: str,
                 canal_password: str, canal_destination: str, canal_client_id: str,
                 binlog_type_name: str, rowtime_from: str, before_column_suffix: str, after_column_suffix: str,
                 update_suffix: str, **kwargs):
        self._mode = canal_mode
        self.bootstrap_servers = bootstrap_servers
        self.topics = topics
        self.canal_execute_time_name = rowtime_from
        self.canal_event_type_name = binlog_type_name
        self.canal_before_column_suffix = before_column_suffix
        self.canal_after_column_suffix = after_column_suffix
        self.canal_update_suffix = update_suffix

        self.canal_table_filter = table_filter if table_filter else ".*\\..*"

        self.canal_host = canal_host
        self.canal_port = int(canal_port)
        self.canal_username = canal_username
        self.canal_destination = canal_destination
        self.canal_password = canal_password
        self.canal_client_id = canal_client_id

    @classmethod
    def build_topics(cls, connection: Connection, session: Session) -> Dict[Tuple[str, str, str], str]:
        topics = dict()
        for version in session.query(ResourceVersion).filter(ResourceVersion.connection_id == connection.id).all():
            if version.name in CANAL_MODE:
                name = version.resource_name
                conn = version.get_connection_connector()
                assert isinstance(conn, dict)
                w_str = f'please check connection {connection.name} config not found topic'
                assert isinstance(conn, dict) and conn.get('topic') is not None, w_str
                topics[name.database, name.name, version.name] = conn['topic']
        return topics

    @classmethod
    def build(cls, pk: Union[str, int]) -> 'Consumer':
        session = DBSession.get_session()
        try:
            query = session.query(Connector)
            if isinstance(pk, int) or pk.isdigit():
                query = query.filter(Connector.id == pk)
            else:
                query = query.filter(Connector.name == pk)

            connector = query.first()
            assert connector, f"Not Found Connector {pk}"
            c_type = connector.type.code
            assert c_type == 'canal', f'connector  {pk} type must be canal current: {c_type}'
            assert connector.target.type.code == 'kafka'
            kafka = connector.target

            topics = cls.build_topics(kafka, session)
            return Consumer(connector.connector_mode, kafka.url, topics, **connector.get_config_parser()['canal'])
        finally:
            session.close()

    from canal.protocol.EntryProtocol_pb2 import RowData, EventType, Column

    SUPPORT_TYPE = {
        EventType.INSERT: CanalMode.insert,
        EventType.DELETE: CanalMode.delete,
        EventType.UPDATE: CanalMode.update,
    }

    @classmethod
    def _convert_type(cls, v: Column) -> any:
        val = v.value
        typ = v.sqlType
        if val is None or len(val) == 0:
            return None
        # type id from java.sql.Types.java
        if typ in {-7, -6, 5, 4, -5, 16}:  # int bool 16 also parse int
            return int(val)
        elif typ in {6, 7, 8, 2, 3}:  # float
            return float(val)
        elif typ == 91:  # DATE 91
            return val
        elif typ == 92 and val is not None:  # TIME 92
            return val + "Z"
        elif typ == 93 and val is not None:  # TIMESTAMP 93
            return val.replace(' ', 'T') + 'Z'
        return val

    def _generate_by_columns(self, columns: list, name_suffix: Optional[str] = None) -> dict:
        return {(x.name if name_suffix is None else x.name + name_suffix): self._convert_type(x) for x in columns}

    def _convert_after_column(self, value: any, updated: bool) -> any:
        if updated:
            return self._convert_type(value)
        return None

    def _generate_after_columns(self, columns: list):
        return {(x.name + self.canal_after_column_suffix): self._convert_after_column(x, x.updated) for x in
                columns if x.updated}

    def _generate_notice(self, event_type: int, row: RowData, execute_time: str) -> bytes:
        from canal.protocol.EntryProtocol_pb2 import EventType
        _cv = self._convert_type
        execute_time_name = self.canal_execute_time_name
        event_type_name = self.canal_event_type_name
        if self._mode.is_upsert():
            data = self._generate_by_columns(row.beforeColumns)
            after_data = self._generate_by_columns(row.afterColumns)
            data.update(after_data)
            data[event_type_name] = event_type
            assert event_type in (EventType.DELETE, EventType.INSERT, EventType.UPDATE)
        elif event_type == EventType.DELETE:
            data = self._generate_by_columns(row.beforeColumns)
        elif event_type == EventType.INSERT:
            data = self._generate_by_columns(row.afterColumns)
        elif event_type == EventType.UPDATE:
            update_suffix = self.canal_update_suffix
            data = {x.name + update_suffix: x.updated for x in row.afterColumns}
            data = self._generate_by_columns(row.beforeColumns, self.canal_before_column_suffix)
            after_data = self._generate_after_columns(row.afterColumns)
            data.update(data)
            data.update(after_data)
        else:
            raise Exception("Not support Now")
        assert execute_time not in data
        data[execute_time_name] = execute_time
        return json.dumps(data, ensure_ascii=False).encode('utf-8')

    def _generate_topic_name(self, database: str, table: str, event_type: str) -> Optional[str]:
        name = database, table, event_type
        if name in self.topics:
            return self.topics[name]

    @classmethod
    def _convert_utc_time(cls, timestamp: int) -> str:
        date = datetime.fromtimestamp(timestamp / 1000)
        # three way for parse timestamp
        # s = date.strftime("%FT%H:%M:%S.%f")[:-3] + "Z"
        # s = date.isoformat('T', timespec='milliseconds') + 'Z'
        s = date.strftime("%FT%H:%M:%S.%f") + "Z"
        return s

    def run_forever(self, client, producer):
        from canal.protocol import EntryProtocol_pb2
        from canal.protocol.EntryProtocol_pb2 import EntryType
        print(datetime.now(), " start running")
        topics = defaultdict(int)
        sleep_times, send_times = 0, 0
        last_execute_time, add_million_seconds = -1, 0
        while True:
            message = client.get(100)
            entries = message['entries']
            for entry in entries:
                entry_type = entry.entryType
                if entry_type in (EntryType.TRANSACTIONBEGIN, EntryType.TRANSACTIONEND):
                    continue
                row_change = EntryProtocol_pb2.RowChange()
                row_change.MergeFromString(entry.storeValue)
                header = entry.header
                database = header.schemaName
                table = header.tableName
                event_type = header.eventType

                # try add million second when meet same execute time
                fix_execute_time = header.executeTime
                if last_execute_time == header.executeTime:
                    if add_million_seconds < 999:
                        add_million_seconds += 1
                        fix_execute_time = header.executeTime + add_million_seconds
                    else:
                        print('over 1000 event in one seconds')
                        fix_execute_time = header.executeTime + add_million_seconds
                else:
                    last_execute_time = header.executeTime
                    add_million_seconds = 0

                row_time = self._convert_utc_time(fix_execute_time)

                logging.debug(' '.join(str(x) for x in
                                       [row_time, fix_execute_time, header.executeTime, header.logfileOffset,
                                        add_million_seconds]))
                for row in row_change.rowDatas:
                    msg = self._generate_notice(event_type, row, row_time)
                    event_type_name = self.SUPPORT_TYPE[event_type]
                    topic_name = self._generate_topic_name(database, table, event_type_name)
                    if topic_name is None or not self._mode.is_support(event_type_name):
                        print('filter {} {} {} - topic{}'.format(database, table, event_type_name, topic_name))
                        continue

                    if topic_name not in topics:
                        print(topic_name, 'get')
                    topics[topic_name] += 1
                    producer.send(topic_name, value=msg)
                    send_times += 1
            if not entries:
                time.sleep(WAIT_TIMES)
                if sleep_times % PRINT_EACH == 0:
                    for t, n in topics.items():
                        print(' Topic: ', t, ' send ', n)
                    topics = defaultdict(int)
                    logger.debug(
                        "{} wait for in {} already send {}".format(datetime.now(), sleep_times, send_times))
                sleep_times += 1

    def run(self):
        from kafka import KafkaProducer
        from canal.client import Client
        client = Client()
        producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers.split(','))

        try:

            client.connect(host=self.canal_host, port=self.canal_port)
            client.check_valid(username=self.canal_username.encode(), password=self.canal_password.encode())
            client.subscribe(client_id=self.canal_client_id.encode(), destination=self.canal_destination.encode(),
                             filter=self.canal_table_filter.encode())
            self.run_forever(client, producer)
        finally:
            client.disconnect()
            producer.close()



# -*- coding: utf-8 -*-
import os
import re
import subprocess
import tempfile
import yaml
from typing import Optional
from jinja2 import Template
from terminado.management import NamedTermManager
from fsqlfly.settings import FSQLFLY_UPLOAD_DIR, FSQLFLY_FLINK_BIN, logger
from fsqlfly.db_helper import Transform, DBDao
from fsqlfly import settings
from fsqlfly.utils.strings import get_job_header, dump_yaml
from fsqlfly.utils.template import generate_template_context


def _create_config(require: str, config: Optional[str], args: dict) -> str:
    tables = []
    catalogs = []
    require = require.strip() if require and require.strip() else ''
    if require:
        tables.extend(DBDao.get_require_table(require))
        catalogs.extend(DBDao.get_require_catalog(require))

    base_config = yaml.load(handle_template(config, args), yaml.FullLoader) if config else dict()
    if base_config is None:
        base_config = dict()
    if base_config.get('tables'):
        base_config['tables'].extend(tables)
    else:
        base_config['tables'] = tables
    base_config['functions'] = DBDao.get_require_functions()
    if base_config.get('catalogs'):
        base_config['catalogs'].extend(catalogs)
    else:
        base_config['catalogs'] = catalogs
    return dump_yaml(base_config)


def _clean(s: str) -> str:
    return re.sub(r'\[\d+(;\d+)?m?', '', s)


def handle_template(text: Optional[str], args: dict) -> str:
    return Template(text).render(**generate_template_context(**args)) if text else ''


def run_transform(transform: Transform, **kwargs) -> (bool, str):
    _, yaml_f = tempfile.mkstemp(suffix='.yaml')
    _, sql_f = tempfile.mkstemp(suffix='.sql')

    yaml_conf = _create_config(require=transform.require, config=transform.yaml, args=kwargs)
    sql = handle_template(transform.sql, kwargs)
    print(yaml_conf, file=open(yaml_f, 'w'))
    print(sql, file=open(sql_f, 'w'))
    print('q\nexit;', file=open(sql_f, 'a+'))
    run_commands = [FSQLFLY_FLINK_BIN, 'embedded',
                    '-s', get_job_header(transform, **kwargs),
                    '--environment', yaml_f,
                    *DBDao.get_require_jar(),
                    '<', sql_f]
    print(' '.join(run_commands))
    try:
        out = subprocess.check_output(' '.join(run_commands), shell=True, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as error:
        return False, "sql:\n {} \n\noutput: \n{} \n\n error: {}\n\nyaml: \n{}".format(transform.sql,
                                                                                       error.stdout.decode(),
                                                                                       error.stderr.decode(),
                                                                                       yaml_conf)
    except Exception as e:
        print(e)
        return False, str(e)

    out_w = _clean(out.decode())

    os.remove(yaml_f)
    os.remove(sql_f)

    print(out_w, file=open('out.shell.txt', 'w'))
    print(out_w)
    return True, out_w


def run_debug_transform(data: dict, manager: NamedTermManager) -> (str, str):
    _, yaml_f = tempfile.mkstemp(suffix='.yaml')
    yaml_conf = _create_config(data.get('require', ''), data.get('yaml', ''), dict())
    print(yaml_conf, file=open(yaml_f, 'w'))
    real_sql = handle_template(data.get('sql', ''), dict())
    name = manager._next_available_name()
    run_commands = [FSQLFLY_FLINK_BIN, 'embedded',
                    '-s', '{}{}'.format(settings.TEMP_TERMINAL_HEAD, str(name)),
                    '--environment', yaml_f,
                    *DBDao.get_require_jar()]
    logger.debug('running commands is : {}'.format(' '.join(run_commands)))
    term = manager.new_terminal(shell_command=run_commands)

    logger.debug('sql :{}'.format(real_sql))
    term.ptyproc.write(real_sql)
    term.term_name = name
    setattr(term, settings.TERMINAL_OPEN_NAME, True)
    term.run_command = ' '.join(run_commands)
    manager.terminals[name] = term
    return name



import os
import hashlib
import logging
import logzero
from logzero import logger
from os.path import join
from dotenv import load_dotenv
from pathlib import Path
from terminado.management import NamedTermManager
from sqlalchemy import create_engine


def generate_cookie_secret(s: str, typ: str = '___cookie_secret') -> str:
    md5 = hashlib.md5()
    md5.update(f'{s}{typ}'.encode())
    return md5.hexdigest()


BASE_DIR = os.path.expanduser('~')
ENV_FILE_PATH_FROM_LOCAL_ENV = os.environ.get('FSQLFLY')
ENV_FILE_PATH = join(BASE_DIR, '.fsqlfly')
if ENV_FILE_PATH_FROM_LOCAL_ENV and os.path.exists(ENV_FILE_PATH_FROM_LOCAL_ENV):
    logger.debug('Change Env File to {}'.format(ENV_FILE_PATH_FROM_LOCAL_ENV))
    ENV_FILE_PATH = ENV_FILE_PATH_FROM_LOCAL_ENV
if os.path.exists(ENV_FILE_PATH) and os.path.isfile(ENV_FILE_PATH):
    logging.debug("Load Env File From {} ".format(ENV_FILE_PATH))
    load_dotenv(dotenv_path=Path(ENV_FILE_PATH))
else:
    logging.debug("Not Found Valid Env File ".format(ENV_FILE_PATH))

ENV = os.environ.get
FSQLFLY_DEBUG = ENV('FSQLFLY_DEBUG') is not None
FSQLFLY_SAVE_MODE_DISABLE = ENV('FSQLFLY_SAVE_MODE_DISABLE') is not None
if FSQLFLY_DEBUG:
    logzero.loglevel(logging.DEBUG)

FSQLFLY_PASSWORD = ENV('FSQLFLY_PASSWORD', 'password')
logger.debug('FSQLFLY_PASSWORD is {}'.format(FSQLFLY_PASSWORD))
FSQLFLY_TOKEN = generate_cookie_secret(FSQLFLY_PASSWORD, '')
FSQLFLY_TOKEN_BYTE = FSQLFLY_TOKEN.encode()
logger.debug('FSQLFLY_TOKEN_BYTE is {}'.format(FSQLFLY_TOKEN_BYTE))

FSQLFLY_COOKIE_SECRET = generate_cookie_secret(FSQLFLY_PASSWORD)

FSQLFLY_DB_URL = ENV('FSQLFLY_DB_URL')

assert FSQLFLY_DB_URL, 'FSQLFLY_DB_URL must not be null'
ENGINE = create_engine(FSQLFLY_DB_URL, echo=FSQLFLY_DEBUG)

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
FSQLFLY_UPLOAD_DIR = ENV('FSQLFLY_UPLOAD_DIR', join(os.path.expanduser('~'), '.fsqlfly_upload'))
if not os.path.exists(FSQLFLY_UPLOAD_DIR):
    logger.debug('Create Upload Base Dir {}'.format(FSQLFLY_UPLOAD_DIR))
    os.makedirs(FSQLFLY_UPLOAD_DIR)
UPLOAD_ROOT_DIR = join(FSQLFLY_UPLOAD_DIR, 'upload')

FSQLFLY_STATIC_ROOT = ENV('FSQLFLY_STATIC_ROOT', join(ROOT_DIR, 'static'))
FSQLFLY_FINK_HOST = ENV('FSQLFLY_FINK_HOST', 'http://localhost:8081')
FSQLFLY_JOB_DAEMON_FREQUENCY = int(ENV('FSQLFLY_JOB_DAEMON_FREQUENCY', '30'))
FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY = int(ENV('FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY', '3'))

assert os.path.exists(FSQLFLY_STATIC_ROOT), "FSQLFLY_STATIC_ROOT ({}) not set correct".format(FSQLFLY_STATIC_ROOT)
INDEX_HTML_PATH = join(FSQLFLY_STATIC_ROOT, 'index.html')
assert os.path.exists(INDEX_HTML_PATH), "FSQLFLY_STATIC_ROOT ({}) can't find index.html".format(FSQLFLY_STATIC_ROOT)
INDEX_HTML = open(INDEX_HTML_PATH, 'rb').read()

FSQLFLY_FLINK_BIN_DIR = ENV('FSQLFLY_FLINK_BIN_DIR', '/opt/flink/bin')
assert os.path.exists(FSQLFLY_FLINK_BIN_DIR), "FSQLFLY_FLINK_BIN_DIR ({}) not exists, please confirm".format(
    FSQLFLY_FLINK_BIN_DIR)
FSQLFLY_FLINK_BIN = join(FSQLFLY_FLINK_BIN_DIR, 'sql-client.sh')

FSQLFLY_FLINK_MAX_TERMINAL = int(ENV('FSQLFLY_FLINK_MAX_TERMINAL', '100'))
FSQLFLY_WEB_PORT = int(ENV('FSQLFLY_WEB_PORT', '8082'))

TERMINAL_MANAGER = NamedTermManager(
    shell_command=[FSQLFLY_FLINK_BIN, 'embedded'],
    max_terminals=FSQLFLY_FLINK_MAX_TERMINAL)

TERMINAL_OPEN_NAME = '__NEED_OPEN_'
FSQLFLY_JOB_LOG_DIR = ENV('FSQLFLY_JOB_LOG_DIR', '/tmp/fsqlfly_job_log')
FSQLFLY_JOB_LOG_FILE = join(FSQLFLY_JOB_LOG_DIR, 'job_damon.log')
os.makedirs(FSQLFLY_JOB_LOG_DIR, exist_ok=True)

TEMP_TERMINAL_HEAD = '0__TEMPORARY__'

# email

FSQLFLY_MAIL_ENABLE = ENV('FSQLFLY_MAIL_ENABLE') is not None
FSQLFLY_MAIL_HOST = ENV('FSQLFLY_MAIL_HOST')
FSQLFLY_MAIL_USER = ENV('FSQLFLY_MAIL_USER')
FSQLFLY_MAIL_PASSWORD = ENV('FSQLFLY_MAIL_PASSWORD')
FSQLFLY_MAIL_RECEIVERS = ENV('FSQLFLY_MAIL_RECEIVERS')



# -*- coding:utf-8 -*-
import json
import attr
from logzero import logger
from typing import Any, Optional, Awaitable
from datetime import datetime, date
from tornado.web import RequestHandler
from fsqlfly import settings
from fsqlfly.common import DBRes
from fsqlfly.utils.strings import dict2camel, dict2underline


def json_obj_hook(o):
    if isinstance(o, (date, datetime)):
        return str(o)


class BaseHandler(RequestHandler):
    """Request handler where requests and responses speak JSON."""

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        return super(BaseHandler, self).data_received(chunk)

    @property
    def terminal_manager(self):
        return self.settings['terminal_manager']

    def _login_by_header(self):
        token = self.request.headers.get('Token')
        if token is not None:
            logger.debug("try login by request header {}".format(token))
            if token == settings.FSQLFLY_TOKEN:
                logger.debug("login success by request header {}".format(token))
                return token
        return None

    def get_current_user(self) -> Any:
        cookie = self.get_cookie('user')
        if cookie:
            return cookie
        header_token = self._login_by_header()
        if header_token:
            return header_token
        token = self.request.arguments.get('token')
        if token is not None and len(token) == 1 and token[0] == settings.FSQLFLY_TOKEN_BYTE:
            logger.debug("login success By token {}".format(token))
            return token

    def get_login_url(self) -> str:
        return '/login'

    def set_login_status(self):
        self.set_secure_cookie('user', 'admin')

    def set_logout_status(self):
        self.clear_cookie('user')

    @property
    def json_body(self) -> dict:
        _name = '_save_json_data'
        if not hasattr(self, _name):
            try:
                json_data = json.loads(self.request.body)
                if isinstance(json_data, dict):
                    json_data = dict2underline(json_data)
                setattr(self, _name, json_data)
            except ValueError:
                setattr(self, _name, {})

        return getattr(self, _name)

    def set_default_headers(self):
        self.set_header('Content-Type', 'application/json; charset=utf-8')

    def write_json(self, res: dict):
        output = json.dumps(res, ensure_ascii=False, default=json_obj_hook)
        self.write(output)
        self.finish()

    def write_res(self, res: DBRes):
        data = attr.asdict(res)
        j_data = json.dumps(dict2camel(data), ensure_ascii=False, default=json_obj_hook)
        self.write(j_data)
        self.finish()



from fsqlfly.db_models import *
import os
import traceback
import yaml
from functools import wraps, partial
from copy import deepcopy
from typing import Callable, Type, Optional, List, Union, Any, TypeVar
from sqlalchemy import and_, event, or_, subquery
from sqlalchemy.engine import Engine
from fsqlfly import settings
from fsqlfly.common import DBRes
from sqlalchemy.orm.session import Session, sessionmaker, query as session_query
from fsqlfly.settings import FSQLFLY_UPLOAD_DIR, FSQLFLY_FLINK_BIN, logger

Query = session_query.Query

SUPPORT_MODELS = {
    'connection': Connection,
    'connector': Connector,
    'schema': SchemaEvent,
    'name': ResourceName,
    'template': ResourceTemplate,
    'version': ResourceVersion,
    'file': FileResource,
    'functions': Functions,
    'transform': Transform,
    'namespace': Namespace,
    'savepoint': TransformSavepoint

}

DBT = TypeVar('DBT', Connection, Connector, SchemaEvent, ResourceName, ResourceTemplate,
              ResourceVersion, FileResource, Functions, Transform, Namespace, TransformSavepoint)


class DBSession:
    engine = None
    _Session = None

    @classmethod
    def init_engine(cls, engine: Engine):
        cls.engine = engine
        cls._Session = sessionmaker(bind=engine)

    @classmethod
    def get_session(cls, *args, **kwargs) -> Session:
        assert cls.engine is not None
        return cls._Session(*args, **kwargs)


def session_add(func: Callable) -> Callable:
    @wraps(func)
    def _add_session(*args, **kwargs):
        session = kwargs['session'] if 'session' in kwargs else DBSession.get_session()
        new_kwargs = {k: v for k, v in kwargs.items() if k != 'session'}
        try:
            res = func(*args, session=session, **new_kwargs)
            session.commit()
            return res
        except Exception as error:
            session.rollback()
            if settings.FSQLFLY_DEBUG:
                raise error
            err = traceback.format_exc()
            return DBRes.sever_error(msg=f'meet {err}')
        finally:
            if 'session' not in kwargs:
                session.close()

    return _add_session


def filter_not_support(func: Callable) -> Callable:
    @wraps(func)
    def _call_(*args, **kwargs):
        model = kwargs['model'] if 'model' in kwargs else args[1]
        if model not in SUPPORT_MODELS:
            return DBRes.api_error(msg=f'{model} not support')
        base = SUPPORT_MODELS[model]
        return func(*args, base=base, **kwargs)

    return _call_


class DBDao:
    @classmethod
    @session_add
    @filter_not_support
    def name2pk(cls, model: str, name: str, *args, session: Session, base: Type[DBT], **kwargs) -> int:
        return session.query(base.id).filter(base.name == name).one()[0]

    @classmethod
    def is_null_foreign_key(cls, k: str, v: Any) -> bool:
        return k.endswith('_id') and v == 0

    @classmethod
    @session_add
    @filter_not_support
    def update(cls, model: str, pk: int, obj: dict, *args, session: Session, base: Type[Base], **kwargs) -> DBRes:
        assert session is not None
        first = session.query(base).filter(base.id == pk).first()
        if first is None:
            return DBRes.not_found()

        if first.is_locked and obj.get('is_locked') is True:
            return DBRes.resource_locked()

        for k, v in obj.items():
            if k not in ['id', 'create_at', 'update_at'] and not cls.is_null_foreign_key(k, v):
                setattr(first, k, v)
        return DBRes(data=first.as_dict())

    @classmethod
    @session_add
    @filter_not_support
    def create(cls, model: str, obj: dict, *args, session: Session, base: Type[DBT], **kwargs) -> DBRes:
        db_obj = base(**obj)
        session.add(db_obj)
        session.commit()
        return DBRes(data=db_obj.as_dict())

    @classmethod
    def build_and(cls, data: dict, base: Type, query: Query) -> Query:
        kvs = [getattr(base, k) == v for k, v in data.items()]
        if len(kvs) > 1:
            query = query.filter(and_(*kvs))
        else:
            query = query.filter(kvs[0])
        return query

    @classmethod
    @session_add
    @filter_not_support
    def get(cls, model: str, *args, session: Session, base: Type[Base], filter_: Optional[dict] = None,
            **kwargs) -> DBRes:
        query = session.query(base)
        if filter_:
            query = cls.build_and(filter_, base, query)
        return DBRes(data=[x.as_dict() for x in query.all()])

    @classmethod
    def get_job_names(cls, *args, session: Session, **kwargs) -> dict:
        res = dict()
        query = session.query(Transform).join(Transform.namespace, isouter=True)
        for q in (and_(Transform.is_daemon.is_(True), Namespace.is_daemon.is_(True)),
                  and_(Transform.namespace_id.is_(None), Transform.is_daemon.is_(True))):
            for x in query.filter(q).all():
                res["{}_{}".format(x.id, x.name)] = x

        return res

    @classmethod
    @session_add
    def get_require_name(cls, *args, session: Session, **kwargs) -> DBRes:
        query = session.query(ResourceVersion.full_name).join(ResourceVersion.connection).join(
            ResourceVersion.resource_name)
        version_data = [x[0] for x in
                        query.filter(and_(Connection.is_active == True, ResourceName.is_active == True)).all()]

        v_query = session.query(ResourceVersion).join(ResourceVersion.connection).join(
            ResourceVersion.resource_name)
        default_version_data = list(x.template.full_name for x in v_query.filter(
            and_(Connection.is_active == True, ResourceName.is_active == True,
                 ResourceVersion.is_default == True)).all())
        t_query = session.query(ResourceTemplate).join(ResourceTemplate.connection).join(
            ResourceTemplate.resource_name)
        resource_data = list(x.resource_name.full_name for x in t_query.filter(
            and_(Connection.is_active == True,
                 ResourceName.is_active == True,
                 ResourceTemplate.is_default == True)).all())

        hive_data = [x[0] for x in session.query(Connection.name).filter(Connection.is_active == True,
                                                                         Connection.type == 'hive').all()]

        return DBRes(data=hive_data + version_data + default_version_data + resource_data)

    @classmethod
    def is_hive_table(cls, full_name: str) -> bool:
        return '.' not in full_name

    @classmethod
    @session_add
    def get_require_table(cls, require: str, *args, session: Session, **kwargs) -> List[dict]:
        names = set()
        res = []
        for full_name in require.split(','):
            if cls.is_hive_table(full_name):
                continue
            version = session.query(ResourceVersion).filter(ResourceVersion.full_name == full_name).first()
            if version is None:
                version = session.query(ResourceVersion).join(ResourceVersion.template).filter(
                    and_(ResourceTemplate.full_name == full_name,
                         ResourceVersion.is_default == True)).first()
                if version is None:
                    version = session.query(ResourceVersion).join(ResourceVersion.template).join(
                        ResourceVersion.resource_name).filter(and_(ResourceName.full_name == full_name,
                                                                   ResourceTemplate.is_default == True,
                                                                   ResourceVersion.is_default == True)).first()

            assert version, "Not Found {} in database, try use full name".format(full_name)
            cache = version.generate_version_cache()
            r_name = version.resource_name
            t_name = version.template
            cache['type'] = version.template.type.code
            for name in [r_name.name, r_name.full_name, t_name.full_name]:
                if name not in names:
                    cache['name'] = name.replace('.', '__')
                    names.add(name)
                    break
            if 'name' in cache:
                res.append(cache)
            full_name_cache = deepcopy(cache)
            full_name_cache['name'] = version.full_name.replace('.', '__')
            res.append(full_name_cache)

        return res

    @classmethod
    @session_add
    def get_require_catalog(cls, require: str, *args, session: Session, **kwargs) -> List[dict]:
        names = set()
        res = []
        for full_name in require.split(','):
            if cls.is_hive_table(full_name):
                hive = session.query(Connection).filter(Connection.is_active == True,
                                                        Connection.type == 'hive',
                                                        Connection.name == full_name).one()
                conn = hive.get_connection_connector()
                conn['name'] = hive.name
                if hive.name not in names:
                    res.append(conn)
                    names.add(hive.name)

        return res

    @classmethod
    @session_add
    def get_require_functions(cls, *args, session: Session, **kwargs) -> List[dict]:
        res = []
        for f in session.query(Functions).filter(Functions.is_active == True).all():
            fc = {
                'name': f.name,
                'from': 'class',
                'class': f.class_name
            }
            if f.constructor_config.strip():
                fc['constructor'] = yaml.load(f.constructor_config, yaml.FullLoader)
            res.append(fc)
        return res

    @classmethod
    @session_add
    def get_require_jar(cls, *args, session: Session, **kwargs) -> List[str]:
        res = []
        for f in session.query(Functions).filter(Functions.is_active == True).all():
            r_p = os.path.join(FSQLFLY_UPLOAD_DIR, f.resource.real_path[1:])
            res.append(r_p)

        out = []
        for x in set(res):
            out.extend(['-j', x])
        return out

    @classmethod
    def one(cls, *args, session: Session,
            base: Type[DBT], pk: int,
            **kwargs) -> DBT:
        return session.query(base).filter(base.id == pk).one()

    @classmethod
    @session_add
    @filter_not_support
    def delete(cls, model: str, pk: int, *args, session: Session, base: Type[Base], **kwargs) -> DBRes:
        if settings.FSQLFLY_SAVE_MODE_DISABLE:
            obj = session.query(base).get(pk)
            session.delete(obj)
            return DBRes(data=obj.id)
        else:
            return DBRes.sever_error('Not Support Delete when FSQLFLY_SAVE_MODE_DISABLE not set')

    @classmethod
    def _clean(cls, obj: DBT, session: Session, base: Type[DBT]):
        back = obj.as_dict()
        session.delete(obj)
        session.commit()
        session.add(base(**back))
        session.commit()

    @classmethod
    @session_add
    @filter_not_support
    def clean(cls, model: str, pk: int, *args, session: Session, base: Type[DBT], **kwargs) -> DBRes:
        obj = session.query(base).get(pk)
        if isinstance(obj, Connector):
            back = obj.as_dict()
            source, target = obj.source, obj.target
            session.delete(obj)
            session.commit()
            cls._clean(source, session, Connection)
            cls._clean(target, session, Connection)
            session.add(Connector(**back))
        else:
            if isinstance(obj, Connection):
                s_names = [x[0] for x in
                           session.query(Connector.name).join(Connector.source).filter(Connection.id == pk).all()]
                t_names = [x[0] for x in
                           session.query(Connector.name).join(Connector.target).filter(Connection.id == pk).all()]
                msg = "please clean connector source: {} target: {}".format(','.join(s_names), ','.join(t_names))
                if s_names or t_names:
                    return DBRes.api_error(msg)
            cls._clean(obj, session, base)
        return DBRes()

    @classmethod
    @session_add
    def bulk_insert(cls, data: list, *args, session: Session, **kwargs):
        session.add_all(data)
        return DBRes(data=len(data))

    @classmethod
    @session_add
    def count(cls, base: Type[Base], *args, session: Session, **kwargs) -> int:
        return session.query(base).count()

    @classmethod
    @session_add
    def get_transform(cls, pk: Optional[int] = None, *args, session: Session, **kwargs) -> Union[List[Base], Base]:
        query = session.query(Transform)
        if pk:
            return query.filter(Transform.id == pk).first()
        return query.all()

    @classmethod
    def save(cls, obj: Base, *args, session: Session, **kwargs) -> Base:
        session.add(obj)
        session.commit()
        return obj

    @classmethod
    def upsert_schema_event(cls, obj: SchemaEvent, *args, session: Session, **kwargs) -> (SchemaEvent, bool):
        inserted = True
        query = session.query(SchemaEvent).filter(and_(SchemaEvent.database == obj.database,
                                                       SchemaEvent.name == obj.name,
                                                       SchemaEvent.connection_id == obj.connection_id))
        res = first = query.order_by(SchemaEvent.version.desc()).first()
        if first:
            if first.fields != obj.fields or first.primary_key != obj.primary_key or obj.partitionable != obj.partitionable:
                obj.version = first.version + 1
                obj.father = first
                res = obj
            else:
                inserted = False
                res.info = obj.info
                res.comment = obj.comment
        else:
            res = obj
        session.add(res)
        session.commit()
        return res, inserted

    @classmethod
    def upsert_resource_name(cls, obj: ResourceName, *args, session: Session, **kwargs) -> (ResourceName, bool):
        inserted = False
        query = session.query(ResourceName).filter(and_(ResourceName.full_name == obj.full_name,
                                                        ResourceName.connection_id == obj.connection_id))
        res = first = query.first()
        if first:
            res.latest_schema_id = obj.latest_schema_id
            res.info = obj.info
            res.is_latest = True
            res.schema_version_id = obj.schema_version_id
        else:
            inserted = True
            res = obj
        session.add(res)
        session.commit()
        return res, inserted

    @classmethod
    def upsert_resource_template(cls, obj: ResourceTemplate, *args,
                                 session: Session, **kwargs) -> (ResourceTemplate, bool):
        query = session.query(ResourceTemplate).filter(and_(ResourceTemplate.name == obj.name,
                                                            ResourceTemplate.connection_id == obj.connection_id,
                                                            ResourceTemplate.resource_name_id == obj.resource_name_id))
        inserted = False
        res = first = query.first()
        if first:
            res.config = obj.config
            res.info = obj.info
            res.is_system = obj.is_system
            res.is_default = obj.is_default
            res.full_name = obj.full_name
            res.schema_version_id = obj.schema_version_id
        else:
            inserted = True
            res = obj
        session.add(res)
        session.commit()
        return res, inserted

    @classmethod
    def upsert_resource_version(cls, obj: ResourceVersion, *args, session: Session,
                                **kwargs) -> (ResourceVersion, bool):
        max_version_obj = session.query(ResourceVersion.version).filter(
            ResourceVersion.template_id == obj.template_id).order_by(ResourceVersion.version.desc()).first()
        max_version = max_version_obj[0] if max_version_obj else 0
        query = session.query(ResourceVersion).filter(and_(ResourceVersion.name == obj.name,
                                                           ResourceVersion.template_id == obj.template_id))
        inserted = True
        res = first = query.order_by(ResourceVersion.version.desc()).first()
        if first:
            if first.config != obj.config:
                res.version = max_version + 1
            res.config = obj.config
            res.cache = obj.cache
            res.info = obj.info
            res.cache = obj.cache
            res.schema_version_id = obj.schema_version_id
            inserted = False
        else:
            res = obj
            res.version = max_version + 1
        session.add(res)
        session.commit()
        return res, inserted

    @classmethod
    def upsert_transform(cls, obj: Transform, *args, session: Session,
                         **kwargs) -> (Transform, bool):
        query = session.query(Transform).filter(Transform.name == obj.name)
        inserted = True
        res = first = query.first()
        if first:
            first.sql = obj.sql
            first.connector_id = obj.connector_id
            first.info = obj.info
            first.require = obj.require
            first.yaml = obj.yaml
            first.namespace_id = obj.namespace_id
            first.is_daemon = obj.is_daemon
            inserted = False
        else:
            res = obj
        session.add(res)
        session.commit()
        return res, inserted

    @classmethod
    def create_all_tables(cls):
        create_all_tables(DBSession.engine)

    @classmethod
    def delete_all_tables(cls, force: bool = False):
        delete_all_tables(DBSession.engine, force)


from fsqlfly.settings import ENGINE

DBSession.init_engine(ENGINE)


def update_default_value(mapper, connection, target, father_name):
    if target.is_default:
        father_id = getattr(target, father_name)
        father_id = int(father_id) if isinstance(father_id, str) else father_id
        connection.execute(
            'update %s set is_default = 0 where id <> %d and is_default = 1 and %s = %d' % (
                mapper.local_table.fullname, target.id, father_name, father_id))


for _mode in ['after_insert', 'after_update']:
    for _model, f_n in zip([ResourceTemplate, ResourceVersion], ['resource_name_id', 'template_id']):
        event.listen(_model, _mode, partial(update_default_value, father_name=f_n))



# -*- coding:utf-8 -*-
from fsqlfly.common import *
import logzero
from copy import deepcopy
from typing import Tuple
from sqlalchemy import create_engine, inspect
from fsqlfly.common import BlinkSQLType
from sqlalchemy.sql.sqltypes import TypeDecorator
from fsqlfly.db_helper import ResourceName, ResourceTemplate, ResourceVersion, Connection, SchemaEvent, Transform
from fsqlfly.utils.strings import load_yaml, dump_yaml, get_full_name
from fsqlfly.db_helper import DBSession, DBDao, Session, SUPPORT_MODELS, Connector, DBT


class UpdateStatus:
    def __init__(self, name: str = ''):
        self.schema_inserted = 0
        self.schema_updated = 0
        self.name_inserted = 0
        self.name_updated = 0
        self.template_inserted = 0
        self.template_updated = 0
        self.version_inserted = 0
        self.version_updated = 0
        self.name = name

    def update_schema(self, inserted: bool):
        self.schema_inserted += inserted
        self.schema_updated += not inserted

    def update_resource_name(self, inserted: bool):
        self.name_inserted += inserted
        self.name_updated += not inserted

    def update_template(self, inserted: bool):
        self.template_inserted += inserted
        self.template_updated += not inserted

    def update_version(self, inserted: bool):
        self.version_inserted += inserted
        self.version_updated += not inserted

    @property
    def info(self) -> str:
        header = f"{self.name}:\n\n\n" if self.name else ''
        return header + '\n'.join(['{}: {}'.format(x, getattr(self, x)) for x in dir(self) if x.endswith('ed')])


class BaseManager:
    use_comment = False
    use_primary = False
    renewable = True

    def __init__(self, connection_url: str, table_name_filter: NameFilter, db_type):
        self.connection_url = connection_url
        self.need_tables = table_name_filter
        self.db_type = db_type

    def run(self, target: DBT, session: Optional[Session] = None) -> DBRes:
        if not self.renewable:
            return DBRes.api_error("{} is Not renewable".format(target.name))
        status = UpdateStatus()

        session = DBSession.get_session() if session is None else session
        if isinstance(target, (Connection, ResourceName)):
            if isinstance(target, ResourceName):
                target = target.connection
            self.update_connection(status, session, target)
        elif isinstance(target, ResourceTemplate):
            if target.is_system:
                self.update_template(status, session, target.connection, target.schema_version, target.resource_name)
            else:
                self.update_version(status, session, target.connection, target.schema_version, target.resource_name,
                                    target)
        elif isinstance(target, ResourceVersion):
            if target.is_system:
                self.update_version(status, session, target.connection, target.schema_version, target.resource_name,
                                    target.template)
            else:
                self.update_version_cache(status, session, target)
        else:
            raise NotImplementedError("Not Support {}".format(type(target)))
        return DBRes(data=status.info)

    def update_connection(self, status: UpdateStatus, session: Session, connection: Connection):
        for content in self.update():
            schema = self.generate_schema_event(schema=content, connection=connection)
            schema, i = DBDao.upsert_schema_event(schema, session=session)
            status.update_schema(i)
            self.update_resource_name(status, session, connection, schema)

    def update_resource_name(self, status: UpdateStatus, session: Session, connection: Connection, schema: SchemaEvent):
        resource_name = self.generate_resource_name(connection, schema)
        resource_name, i = DBDao.upsert_resource_name(resource_name, session=session)
        status.update_resource_name(i)
        self.update_template(status, session, schema=schema, connection=connection, resource_name=resource_name)

    def update_template(self, status: UpdateStatus, session: Session, connection: Connection, schema: SchemaEvent,
                        resource_name: ResourceName):
        for template in self.generate_template(schema=schema, connection=connection, resource_name=resource_name):
            template, i = DBDao.upsert_resource_template(template, session=session)
            status.update_template(i)
            self.update_version(status, session, schema=schema, connection=connection, resource_name=resource_name,
                                template=template)

    def update_version(self, status: UpdateStatus, session: Session, connection: Connection, schema: SchemaEvent,
                       resource_name: ResourceName, template: ResourceTemplate):
        version = self.generate_default_version(connection, schema, resource_name, template)
        version, i = DBDao.upsert_resource_version(version, session=session)
        status.update_version(i)
        version.cache = dump_yaml(version.generate_version_cache())
        DBDao.save(version, session=session)

    @classmethod
    def update_version_cache(cls, status: UpdateStatus, session: Session, version: ResourceVersion):
        version.cache = dump_yaml(version.generate_version_cache())
        DBDao.save(version, session=session)
        status.update_version(False)

    @classmethod
    def generate_schema_event(cls, schema: SchemaContent, connection: Connection) -> SchemaEvent:
        return SchemaEvent(name=schema.name, info=schema.comment, database=schema.database,
                           connection_id=connection.id, comment=schema.comment, primary_key=schema.primary_key,
                           fields=dump_yaml([attr.asdict(x) for x in schema.fields]),
                           partitionable=schema.partitionable)

    def update(self) -> List[SchemaContent]:
        return list()

    @classmethod
    def generate_resource_name(cls, connection: Connection, schema: SchemaEvent) -> ResourceName:
        return ResourceName(name=schema.name, database=schema.database, connection_id=connection.id,
                            schema_version_id=schema.id, latest_schema_id=schema.id, is_latest=True,
                            full_name=get_full_name(connection.name, schema.database, schema.name))

    support_type = []

    def generate_template(self, connection: Connection, schema: SchemaEvent,
                          resource_name: ResourceName) -> List[ResourceTemplate]:
        res = []
        for x in self.support_type:
            temp = ResourceTemplate(name=x, type=x, is_system=True,
                                    full_name=get_full_name(resource_name.full_name, x),
                                    connection_id=connection.id, resource_name_id=resource_name.id,
                                    schema_version_id=schema.id)
            res.append(temp)
        return res

    def generate_default_version_config(self, connection: Connection, schema: SchemaEvent, name: ResourceName,
                                        template: ResourceTemplate) -> VersionConfig:
        raise NotImplementedError

    def generate_default_version(self, connection: Connection, schema: SchemaEvent, name: ResourceName,
                                 template: ResourceTemplate) -> ResourceVersion:
        config = self.generate_default_version_config(connection, schema, name, template)

        config_str = dump_yaml(attr.asdict(config))
        return ResourceVersion(name=ResourceVersion.latest_name(),
                               full_name=get_full_name(template.full_name, ResourceVersion.latest_name()),
                               is_system=True, is_latest=True,
                               connection_id=connection.id, resource_name_id=name.id, template_id=template.id,
                               schema_version_id=schema.id if schema else None, config=config_str)


class DatabaseManager(BaseManager):
    support_type = ['sink', 'source', 'both']

    def generate_default_version_config(self, connection: Connection, schema: SchemaEvent, name: ResourceName,
                                        template: ResourceTemplate) -> VersionConfig:
        config = VersionConfig()
        if template.type == 'sink':
            if name.get_config('insert_primary_key', typ=bool) and schema.primary_key:
                config.exclude = schema.primary_key
        return config

    use_comment = True
    use_primary = True

    def _get_flink_type(self, typ: TypeDecorator) -> Optional[str]:
        from sqlalchemy.dialects.mysql.types import _StringType as M_STRING, BIT as M_BIT, TINYINT, DOUBLE as M_DOUBLE
        from sqlalchemy.dialects.postgresql import (DOUBLE_PRECISION as P_DOUBLE, BIT as P_BIT, VARCHAR, CHAR, TEXT,
                                                    JSON, BOOLEAN)
        from sqlalchemy.sql.sqltypes import (DATE, _Binary, INTEGER, SMALLINT, BIGINT, FLOAT, DECIMAL,
                                             DATETIME,
                                             TIMESTAMP, TIME)
        name = None
        types = BlinkSQLType
        if isinstance(typ, (M_STRING, VARCHAR, CHAR, TEXT, JSON)):
            name = types.STRING
        elif isinstance(typ, _Binary):
            name = types.BYTES
        elif isinstance(typ, BOOLEAN):
            name = types.BOOLEAN
        elif isinstance(typ, TINYINT):
            name = types.INTEGER
            if typ.display_width == 1:
                name = types.BOOLEAN
        elif isinstance(typ, SMALLINT):
            name = types.SMALLINT
        elif isinstance(typ, (P_BIT, M_BIT)):
            name = types.TINYINT
        elif isinstance(typ, INTEGER):
            name = types.INTEGER
        elif isinstance(typ, BIGINT):
            name = types.BIGINT
        elif isinstance(typ, FLOAT):
            name = types.FLOAT
        elif isinstance(typ, (M_DOUBLE, P_DOUBLE)):
            name = types.DOUBLE
        elif isinstance(typ, DECIMAL):
            name = types.DECIMAL
        elif isinstance(typ, DATETIME) or isinstance(typ, TIMESTAMP):
            name = types.TIMESTAMP
        elif isinstance(typ, DATE):
            name = types.DATE
        elif isinstance(typ, TIME):
            name = types.TIME

        if name is None:
            logzero.logger.error("Not Support Current Type in DB {}".format(str(typ)))
            return None
        return name

    def update(self):
        update_cache_name = '__update_cache_name'
        if hasattr(self, update_cache_name):
            return getattr(self, update_cache_name)
        from sqlalchemy.sql.sqltypes import INTEGER, SMALLINT, BIGINT
        engine = create_engine(self.connection_url)
        insp = inspect(engine)
        db_list = insp.get_schema_names()
        update_tables = []
        for db in db_list:
            for tb in insp.get_table_names(db):
                full_name = f'{db}.{tb}'
                if full_name in self.need_tables:
                    update_tables.append((db, tb))

        schemas = []

        for db, tb in update_tables:
            comment = None
            if self.use_comment:
                try:
                    comment = insp.get_table_comment(table_name=tb, schema=db)['text']
                except NotImplementedError as err:
                    print('meet ', err)

            schema = SchemaContent(name=tb, database=db, comment=comment, type=self.db_type)
            columns = insp.get_columns(table_name=tb, schema=db)
            primary = insp.get_primary_keys(tb, db) if self.use_primary else None
            if primary and len(primary) == 1:
                schema.primary_key = primary[0]
                column = list(filter(lambda x: x['name'] == primary[0], columns))[0]
                if isinstance(column['type'], (INTEGER, SMALLINT, BIGINT)):
                    schema.partitionable = True

            fields = []
            for x in columns:
                field = SchemaField(name=x['name'], type=self._get_flink_type(x['type']),
                                    nullable=x['nullable'], autoincrement=x.get('autoincrement'))
                if field.type is None:
                    logzero.logger.error(
                        "Not Add Column {} in {}.{} current not support : {}".format(field.name, schema.database,
                                                                                     schema.name, str(x['type'])))
                else:
                    fields.append(field)

            schema.fields.extend(fields)

            # print(schema)
            schemas.append(schema)
        setattr(self, update_cache_name, schemas)
        return schemas


class HiveManager(DatabaseManager):
    use_comment = False
    use_primary = False
    support_type = []

    hive_types = set(list(x for x in dir(BlinkHiveSQLType) if not x.startswith('__')))

    def _get_flink_type(self, typ: TypeDecorator) -> Optional[str]:
        from sqlalchemy.sql.sqltypes import (VARCHAR, String, CHAR, BINARY)

        name = None
        types = BlinkHiveSQLType
        str_name = str(typ)

        if isinstance(typ, (VARCHAR, CHAR, String)):
            name = types.STRING
        elif isinstance(typ, BINARY):
            name = types.BYTES
        elif str_name in self.hive_types:
            name = str_name
        if name is None:
            logzero.logger.error("Not Support Current Type in DB {}".format(str(typ)))
            return None
        return name


class ElasticSearchManager(DatabaseManager):
    def _es2flink(self, typ: str) -> Optional[str]:
        types = BlinkSQLType
        if typ in ("text", "keyword"):
            return types.STRING
        elif typ == 'long':
            return types.BIGINT
        elif typ == 'integer':
            return types.INTEGER
        elif typ == 'byte':
            return types.TINYINT
        elif typ == 'short':
            return types.SMALLINT
        elif typ == 'float':
            return types.FLOAT
        elif typ == 'binary':
            return types.BYTES
        elif typ == 'boolean':
            return types.BOOLEAN
        elif typ == 'date':
            return types.DATE

    def update(self):
        assert 'elasticsearch' == self.db_type
        from elasticsearch import Elasticsearch
        es = Elasticsearch(hosts=self.connection_url)
        schemas = []
        for index in es.indices.get_alias('*'):
            if index not in self.need_tables:
                continue

            schema = SchemaContent(name=index, type=self.db_type)
            mappings = es.indices.get_mapping(index)

            fields = []
            for k, v in mappings[index]['mappings']['properties'].items():
                field = SchemaField(name=k, type=self._es2flink(v['type']),
                                    nullable=True)
                if field.type is None:
                    logzero.logger.error(
                        "Not Add Column {} in {}.{} current not support : {}".format(field.name, schema.database,
                                                                                     schema.name, str(v['type'])))
                else:
                    fields.append(field)

            fields.sort(key=lambda x: x.name)

            schema.fields.extend(fields)

            print(schema)
            schemas.append(schema)

        return schemas


class HBaseManager(DatabaseManager):
    def update(self):
        assert 'hbase' == self.db_type
        import happybase
        host, port = self.connection_url.split(':')
        connection = happybase.Connection(host, int(port), autoconnect=True)
        schemas = []
        for x in connection.tables():
            tab = x.decode()
            table = connection.table(tab)
            schema = SchemaContent(name=tab, type=self.db_type)
            fields = []
            for fm in table.families():
                fields.append(SchemaField(name=fm.decode(), type=BlinkHiveSQLType.BYTES, nullable=True))
            schema.fields.extend(fields)
            schemas.append(schema)
        return schemas


class FileManger(BaseManager):
    renewable = False


class KafkaManger(BaseManager):
    renewable = False


SUPPORT_MANAGER = {
    'hive': HiveManager,
    'jdbc': DatabaseManager,
    'kafka': KafkaManger,
    'hbase': HBaseManager,
    'elasticsearch': ElasticSearchManager,
    'filesystem': FileManger
}


class ConnectorManager:
    def _run(self, connector: Connector, session: Session) -> DBRes:
        raise NotImplementedError

    def run(self, connector: Connector, _session: Optional[Session] = None) -> DBRes:
        session = DBSession.get_session() if _session is None else _session
        try:
            return self._run(connector, session)
        finally:
            session.close()

    @classmethod
    def check_system(cls, connector: Connector):
        s_type, t_type = connector.source.type.code, connector.target.type.code
        msg = 'system only support db -> hive: current {}->{}'.format(s_type, t_type)
        assert s_type == 'jdbc' and t_type == 'hive', msg


class CanalKafkaManager(DatabaseManager):
    support_type = ['source']

    renewable = True

    def __init__(self, connector: Connector, mode: str, cache: List[SchemaContent]):
        super().__init__(connector.target.url, NameFilter(), 'kafka')
        self._cache = cache
        self._connector = connector
        self._mode = mode

    def update(self) -> List[SchemaContent]:
        return self._cache

    def generate_default_version(self, connection: Connection, schema: SchemaEvent, name: ResourceName,
                                 template: ResourceTemplate) -> ResourceVersion:
        config = super(CanalKafkaManager, self).generate_default_version_config(connection, schema, name, template)
        version = super(CanalKafkaManager, self).generate_default_version(connection, schema, name, template)
        mode = self._mode
        version.name = mode
        version.full_name = get_full_name(template.full_name, mode)

        schema_fields = [SchemaField(**x) for x in load_yaml(schema.fields)] if schema else []
        row_field = self._connector.rowtime_field
        process_field = self._connector.process_time_field
        execute_time_field = self._connector.db_execute_time_field
        binlog_field = self._connector.binlog_type_name_field
        cnt = self._connector
        fields = []

        if mode == CanalMode.update:
            config.exclude = '.*'
            for schema in schema_fields:
                for suffix, tp in zip([cnt.before_column_suffix, cnt.after_column_suffix, cnt.update_suffix],
                                      [schema.type, schema.type, BlinkSQLType.BOOLEAN]):
                    n_schema = deepcopy(schema)
                    n_schema.type = tp
                    n_schema.name = n_schema.name + suffix
                    fields.append(n_schema)

        if row_field:
            fields.append(row_field)
        if process_field:
            fields.append(process_field)

        if self._mode == CanalMode.upsert:
            fields.append(binlog_field)

        if execute_time_field:
            fields.append(execute_time_field)

        config.schema = [ResourceVersion.field2schema(x) for x in fields]

        version.config = dump_yaml(attr.asdict(config))
        return version


class CanalManager:

    def __init__(self, connector: Connector, session: Session):
        db, kafka = connector.source, connector.target
        self.connector = connector
        self.db = db
        self.kafka = kafka
        self.session = session
        self.db_manager = DatabaseManager(db.url, NameFilter(db.include, db.exclude), db.type)
        self.mode = self.connector.connector_mode

    def run(self) -> DBRes:
        db, kafka, session, db_manager = self.db, self.kafka, self.session, self.db_manager
        res = []
        db_status = UpdateStatus('MySQL Status')
        db_manager.update_connection(db_status, session, db)
        res.append(db_status)
        for mode in self.mode:
            manager = CanalKafkaManager(self.connector, mode, db_manager.update())
            status = UpdateStatus('Canal {} Status'.format(mode))
            manager.update_connection(status, session, kafka)
            res.append(status)

        return DBRes(msg='\n\n'.join(x.info for x in res))


class CanalConnectorManager(ConnectorManager):
    def _run(self, connector: Connector, session: Session) -> DBRes:
        s_type, t_type = connector.source.type.code, connector.target.type.code
        assert s_type == 'jdbc', 'Canal Source Type must be db current: {}'.format(s_type)
        assert t_type == 'kafka', 'Canal Target Type must be kafka current: {}'.format(t_type)

        manager = CanalManager(connector, session)

        return manager.run()


class SystemConnectorManager(ConnectorManager):
    def _run(self, connector: Connector, session: Session) -> DBRes:
        self.check_system(connector)
        need_tables = connector.need_tables
        source, target = connector.source, connector.target
        if not connector.source.resource_names:
            return DBRes.api_error("Not Found Any Resource Name in Connection {}".format(source.name))
        resource_names = [x for x in connector.source.resource_names if x.db_name in need_tables]
        return self.handle(resource_names, connector, session)


class SystemConnectorUpdateManager(SystemConnectorManager):
    @classmethod
    def build_sql(cls, target_database: str, target_table: str, version: ResourceVersion, connector: Connector) -> str:
        schemas = version.generate_version_cache()['schema']
        table = f"{connector.target.name}.{target_database}.{target_table}"
        source = version.db_full_name
        fields = ','.join("`{}`".format(x['name']) for x in schemas)
        key, value = connector.partition_key_value
        partition = f"PARTITION ( {key} = '{value}' ) " if connector.use_partition else ''
        way = 'OVERWRITE' if connector.system_overwrite else 'INTO'
        sql = f"INSERT {way} {table} {partition} select {fields} from {source};"

        return sql

    def handle(self, resource_names: List[ResourceName], connector: Connector, session: Session) -> DBRes:
        updated = inserted = 0

        for resource_name in resource_names:
            name = connector.get_transform_name_format(resource_name=resource_name)
            require_version = None
            for version in resource_name.versions:
                if version.template.type == 'source':
                    require_version = version
            assert require_version, 'Not found require version Please check {}'.format(resource_name.full_name)
            require = require_version.full_name + ',' + connector.target.name
            t_database, t_table = connector.get_transform_target_full_name(resource_name=resource_name,
                                                                           connector=connector)
            execution = dict(planner='blink', type='batch', parallelism=connector.system_execution_parallelism)
            transform = Transform(name=name, sql=self.build_sql(t_database, t_table, require_version, connector),
                                  require=require, connector_id=connector.id, yaml=dump_yaml(dict(execution=execution)))
            transform, i = DBDao.upsert_transform(transform, session=session)
            inserted += i
            updated += not i

        msg = 'update: {}\ninserted: {}'.format(updated, inserted)
        return DBRes(msg=msg)


class SystemConnectorInitManager(SystemConnectorManager):
    @classmethod
    def flink2hive(cls, typ: str) -> str:
        if typ == BlinkSQLType.TIMESTAMP:
            return 'TIMESTAMP'
        if typ == BlinkSQLType.BYTES:
            return BlinkSQLType.STRING
        return typ

    @classmethod
    def build_sql(cls, target_database: str, target_table: str, schema: list, connector: Connector) -> List[str]:
        res = []
        drop_table = f'DROP TABLE IF EXISTS `{target_database}`.`{target_table}`'
        create_header = f'CREATE TABLE `{target_database}`.`{target_table}` ('
        cols = []
        for x in schema:
            typ = cls.flink2hive(x['data-type'])
            col = '`{}` {}'.format(x['name'], typ)
            cols.append(col)
        res.extend([create_header])
        res.append(','.join(cols))
        if connector.use_partition:
            key, _ = connector.partition_key_value
            partition = f") PARTITIONED BY ({key} STRING )"
            res.append(partition)
        else:
            res.append(')')
        return [drop_table, '\n'.join(res)]

    def handle(self, resource_names: List[ResourceName], connector: Connector, session: Session) -> DBRes:
        engine = create_engine(connector.target.url)
        for resource_name in resource_names:
            require_version = None
            for version in resource_name.versions:
                if version.template.type == 'source':
                    require_version = version
            assert require_version, 'Not found require version Please check {}'.format(resource_name.full_name)
            t_database, t_table = connector.get_transform_target_full_name(resource_name=resource_name,
                                                                           connector=connector)
            schemas = version.generate_version_cache()['schema']
            for sql in self.build_sql(t_database, t_table, schemas, connector):
                print(sql)
                engine.execute(sql)

        return DBRes()


class SystemConnectorListManager(SystemConnectorManager):
    def handle(self, resource_names: List[ResourceName], connector: Connector, session: Session) -> DBRes:
        res = []
        for resource_name in resource_names:
            name = connector.get_transform_name_format(resource_name=resource_name)
            res.append(name)
        return DBRes(data=res)


class BaseHelper:
    @classmethod
    def is_support(cls, mode: str) -> bool:
        return mode in ('update', 'clean')

    @classmethod
    def update(cls, model: str, pk: int) -> DBRes:
        raise NotImplementedError

    @classmethod
    def clean(cls, model: str, pk: int) -> DBRes:
        return DBDao.clean(model, pk)


class ConnectorHelper(BaseHelper):
    @classmethod
    def is_support(cls, mode: str) -> bool:
        return mode in ('update', 'clean', 'init', 'list')

    @classmethod
    def get_object_from_db(cls, model: str, pk: int) -> Tuple[Session, DBT]:
        session = DBSession.get_session()
        obj = DBDao.one(base=SUPPORT_MODELS[model], pk=pk, session=session)
        return session, obj

    @classmethod
    def get_manager(cls, type_name: str, method: str = 'update') -> ConnectorManager:
        if method == 'update' and type_name == 'canal':
            return CanalConnectorManager()
        else:
            assert type_name == 'system', 'only system support {}'.format(method)
            if method == 'update':
                return SystemConnectorUpdateManager()
            elif method == 'init':
                return SystemConnectorInitManager()
            elif method == 'list':
                return SystemConnectorListManager()
            raise NotImplementedError("Not Support {}".format(method))

    @classmethod
    def _real_run(cls, model: str, pk: int, method: str) -> DBRes:
        session = DBSession.get_session()
        obj = DBDao.one(base=SUPPORT_MODELS[model], pk=pk, session=session)
        manager = cls.get_manager(obj.type.code, method)
        return manager.run(obj, session)

    @classmethod
    def update(cls, model: str, pk: int) -> DBRes:
        assert model == 'connector', 'Only support Connector Model'
        return cls._real_run(model, pk, 'update')

    @classmethod
    def init(cls, model: str, pk: int) -> DBRes:
        return cls._real_run(model, pk, 'init')

    @classmethod
    def list(cls, model: str, pk: int) -> DBRes:
        return cls._real_run(model, pk, 'list')


class ConnectionHelper(BaseHelper):
    @classmethod
    def update(cls, model: str, pk: int):
        session = DBSession.get_session()
        try:
            obj = DBDao.one(base=SUPPORT_MODELS[model], pk=pk, session=session)
            if isinstance(obj, Connection):
                name_filter = NameFilter(obj.include, obj.exclude)
                manager = SUPPORT_MANAGER[obj.type](obj.url, name_filter, obj.type)
            elif isinstance(obj, ResourceName):
                name_filter = NameFilter(obj.get_include())
                typ = obj.connection.type
                manager = SUPPORT_MANAGER[typ](obj.connection.url, name_filter, typ)
            else:
                name_filter = NameFilter()
                typ = obj.connection.type
                manager = SUPPORT_MANAGER[typ](
                    obj.connection.url,
                    name_filter,
                    typ
                )

            return manager.run(obj, session)
        finally:
            session.commit()
            session.close()


class ManagerHelper:
    @classmethod
    def get_helper(cls, model) -> Union[Type[ConnectorHelper], Type[ConnectionHelper]]:
        if model == 'connector':
            return ConnectorHelper
        else:
            return ConnectionHelper

    @classmethod
    def is_support(cls, model: str, mode: str) -> bool:
        return cls.get_helper(model).is_support(mode)

    @classmethod
    def run(cls, model: str, mode: str, pk: Union[str, int]) -> DBRes:
        if model not in SUPPORT_MODELS:
            return DBRes.api_error(msg='{} not support'.format(model))
        if isinstance(pk, str) and not pk.isnumeric():
            pk = DBDao.name2pk(model, name=pk)
        return getattr(cls.get_helper(model), mode)(model, pk if isinstance(pk, int) else int(pk))



# -*- coding:utf-8 -*-
from __future__ import absolute_import, unicode_literals, print_function
import unittest
from unittest.mock import patch, Mock
from fsqlfly.db_helper import Connector
from fsqlfly.connection_manager import *
from sqlalchemy import create_engine, event, func


class NameFilterTest(unittest.TestCase):
    def test_no_args(self):
        flt = NameFilter()
        self.assertTrue('asdf.csdfd' in flt)

    def test_with_args(self):
        flt = NameFilter('.*', 'ab\..*')
        self.assertTrue('abc.eff' in flt)
        self.assertTrue('ab.eff' not in flt)

        flt = NameFilter('ab[cde]?\..*', 'ab\..*')
        self.assertTrue('abc.eff' in flt)
        self.assertTrue('ab.eff' not in flt)
        self.assertTrue('abd.eff' in flt)
        self.assertTrue('abee.eff' not in flt)

        flt = NameFilter('.*\.alltypes')

        self.assertTrue('abc.alltypes' in flt)
        self.assertTrue('abc.lltypes' not in flt)
        self.assertTrue('abc.alltype' not in flt)


class ManagerTest(unittest.TestCase):
    def test_db_manager(self):
        name_filter = NameFilter('.*\.alltypes')

        mn = DatabaseManager('mysql+pymysql://root:password@localhost:3306/fsqlfly', name_filter, 'db')

        mn.update()

    def test_hive_manager(self):
        name_filter = NameFilter('test\.alltypes')
        mn = HiveManager('hive://localhost:10000/default', name_filter, 'hive')

        mn.update()

    def test_elasticsearch_manager(self):
        name_filter = NameFilter()
        mn = ElasticSearchManager('http://localhost:9200', name_filter, 'elasticsearch')

        mn.update()

    def test_hbase_manager(self):
        name_filter = NameFilter()
        mn = HBaseManager('127.0.0.1:9090', name_filter, 'hbase')
        mn.update()

    def setUp(self) -> None:
        self.sqlite_url = 'sqlite:///test.sqlite3'
        engine = create_engine(self.sqlite_url, echo=True)
        event.listen(engine, 'connect', lambda con, _: con.execute('pragma foreign_keys=ON'))
        DBSession.init_engine(engine)
        DBDao.delete_all_tables(True)
        DBDao.create_all_tables()
        self.engine = engine

    def test_run_with_connection(self):
        connection_args = [
            (self.sqlite_url, 'jdbc', '')
        ]
        for url, typ, connector in connection_args:
            connection = Connection(name='a', url=url, type=typ, connector=connector)
            session = DBSession.get_session()
            session.add(connection)
            session.commit()
            res = ManagerHelper.update('connection', str(connection.id))
            self.assertTrue(res.success)
            self.assertTrue(session.query(SchemaEvent).count() > 0)
            self.assertTrue(session.query(ResourceName).count() > 0)
            self.assertTrue(session.query(ResourceTemplate).count() > 0)
            self.assertTrue(session.query(ResourceVersion).count() > 0)
            self.assertTrue(session.query(ResourceVersion).filter(ResourceVersion.cache.isnot(None)).count() > 0)
            session.close()

    def test_update_canal_connector(self):
        from fsqlfly.settings import FSQLFLY_DB_URL
        db = Connection(name='db', url=FSQLFLY_DB_URL, type='jdbc', connector='', include='sample.*')
        k_connector = """
type: kafka
version: universal     # required: valid connector versions are    "0.8", "0.9", "0.10", "0.11", and "universal"
properties:
  zookeeper.connect: localhost:2181  # required: specify the ZooKeeper connection string
  bootstrap.servers: localhost:9092  # required: specify the Kafka server connection string
  group.id: testGroup                # optional: required in Kafka consumer, specify consumer group

topic: {{ resource_name.database }}__{{ resource_name.name }}__{{ version.name }}        
        """
        kafka = Connection(name='kafka', url='localhost:9092', type='kafka', connector=k_connector)
        session = DBSession.get_session()
        session.add_all([db, kafka])
        session.commit()
        config = """[canal]
mode = insert,update
canal_host: localhost
canal_port: 11111
canal_username: root
canal_password: password
canal_client_id: 11021
canal_destination: example
canal_filter: .*\..*        
        """
        connector = Connector(name='example', type='canal', source_id=db.id, target_id=kafka.id, config=config)
        session.add(connector)
        session.commit()

        res = ManagerHelper.update('connector', connector.id)
        self.assertTrue(res.success)

        self.assertTrue(session.query(SchemaEvent).count() > 0)
        res = session.query(ResourceVersion.template_id, func.count(ResourceVersion.id)).filter(
            ResourceVersion.name != 'latest').group_by(ResourceVersion.template_id).all()
        self.assertTrue(all(map(lambda x: x[1] == 2, res)))

        connector_id = connector.id
        session.close()
        from fsqlfly.contrib.canal import Consumer

        # Consumer.build(connector_id).run()


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Union
from datetime import timedelta, datetime
from fsqlfly.utils import macros


def generate_template_context(execution_date: Optional[Union[datetime, str]] = None, **kwargs):
    execution_date = execution_date if execution_date else datetime.now()
    if isinstance(execution_date, str):
        fmt = '%Y-%m-%d' if len(execution_date) == 10 else '%Y-%m-%d %H:%M:%S'
        execution_date = datetime.strptime(execution_date, fmt)
    ds = execution_date.strftime('%Y-%m-%d')
    ts = execution_date.isoformat()
    yesterday_ds = (execution_date - timedelta(1)).strftime('%Y-%m-%d')
    tomorrow_ds = (execution_date + timedelta(1)).strftime('%Y-%m-%d')

    ds_nodash = ds.replace('-', '')
    ts_nodash = execution_date.strftime('%Y%m%dT%H%M%S')
    ts_nodash_with_tz = ts.replace('-', '').replace(':', '')
    yesterday_ds_nodash = yesterday_ds.replace('-', '')
    tomorrow_ds_nodash = tomorrow_ds.replace('-', '')

    return {
        'ds': ds,
        'ds_nodash': ds_nodash,
        'ts': ts,
        'ts_nodash': ts_nodash,
        'ts_nodash_with_tz': ts_nodash_with_tz,
        'yesterday_ds': yesterday_ds,
        'yesterday_ds_nodash': yesterday_ds_nodash,
        'tomorrow_ds': tomorrow_ds,
        'tomorrow_ds_nodash': tomorrow_ds_nodash,
        'END_DATE': ds,
        'end_date': ds,
        'execution_date': execution_date,
        'latest_date': ds,
        'macros': macros,
        **kwargs
    }



import unittest
import random
import sqlalchemy as sa
from sqlalchemy.orm import sessionmaker
from sqlalchemy import event
from fsqlfly.db_helper import *


class FSQLFlyTestCase(unittest.TestCase):
    def _fk_pragma_on_connect(self, dbapi_con, con_record):
        dbapi_con.execute('pragma foreign_keys=ON')

    def setUp(self) -> None:
        engine = sa.create_engine('sqlite://', echo=True)
        event.listen(engine, 'connect', self._fk_pragma_on_connect)

        DBSession.init_engine(engine)
        delete_all_tables(engine, force=True)
        create_all_tables(engine)
        self.DBSession = DBSession
        self.session = self.get_session()
        self.engine = engine

    def get_session(self):
        return self.DBSession.get_session()

    def tearDown(self) -> None:
        self.session.commit()
        self.session.close()


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-
from fsqlfly.common import safe_authenticated
from fsqlfly.base_handle import BaseHandler
from fsqlfly.db_helper import SUPPORT_MODELS, DBDao


class APICounter(BaseHandler):
    @safe_authenticated
    def get(self):
        data = {k + 'Num': DBDao.count(v) for k, v in SUPPORT_MODELS.items()}
        data['code'] = 200
        data['success'] = True
        return self.write_json(data)


class CRHandler(BaseHandler):
    @safe_authenticated
    def get(self, model: str):
        if model == 'require':
            return self.write_res(DBDao.get_require_name())

        def b2v(x):
            v = x[0].decode()
            if v.isdigit():
                return int(v)
            return v

        filter_ = dict({k: b2v(v) for k, v in self.request.arguments.items() if not k.startswith('_')})
        self.write_res(DBDao.get(model, filter_=filter_))

    @safe_authenticated
    def post(self, model: str):
        self.write_res(DBDao.create(model, self.json_body))


class UDHandler(BaseHandler):
    @safe_authenticated
    def post(self, model: str, pk: int):
        self.write_res(DBDao.update(model, pk, self.json_body))

    @safe_authenticated
    def delete(self, model: str, pk: int):
        self.write_res(DBDao.delete(model, pk))


default_handlers = [
    (r'/api/count', APICounter),
    (r'/api/(?P<model>\w+)', CRHandler),
    (r'/api/(?P<model>\w+)/(?P<pk>\d+)', UDHandler),
]



# -*- coding:utf-8 -*-
import time
import re
import math
from typing import List, Any, Set
from collections import namedtuple, Counter
from datetime import datetime
from requests import Session
from logzero import logger
from fsqlfly.settings import FSQLFLY_FINK_HOST
from fsqlfly.workflow import run_transform
from fsqlfly.utils.strings import get_job_short_name
from fsqlfly.common import DBRes
from fsqlfly.db_helper import DBDao, Transform, DBSession

JobStatus = namedtuple('JobStatus', ['name', 'job_id', 'status', 'full_name',
                                     'start_time', 'end_time', 'duration', 'pt'])
FAIL_HEADER = 'FAIL:'
SUCCESS_HEADER = 'SUCCESS:'


class Cache:
    def __init__(self, max_store_seconds=5):
        self.time = dict()
        self.store = dict()
        self._max = max_store_seconds

    def get(self, name: str) -> Any:
        if name not in self.store:
            return None
        assert name in self.time
        if (time.time() - self.time[name]) > self._max:
            del self.time[name]
            del self.store[name]
            return None
        return self.store[name]

    def set(self, name: str, value: Any):
        self.store[name] = value
        self.time[name] = time.time()

    def remove(self, name: str):
        if name in self.time:
            del self.time[name]
        if name in self.store[name]:
            del self.store[name]


class JobControl:
    restart = 'restart'
    stop = 'stop'
    start = 'start'
    cancel = 'cancel'
    status = 'status'
    RUN_STATUS = 'RUNNING'
    FINISHED_STATUS = 'FINISHED'
    FAIL_STATUS = 'FAILED'

    def __contains__(self, item):
        if hasattr(self, item):
            return True
        return False

    def __init__(self, flink_host=None):
        self.host = flink_host
        self.session = Session()
        self.cache = Cache()

    def _get_job_status(self, job_id: str) -> JobStatus:
        def p_time(t: int) -> str:
            return str(datetime.fromtimestamp(t / 1000))[:19] if t > 0 else '-'

        def p_duration(t: int) -> str:
            second = 1000
            minute = second * 60
            hour = minute * 60
            day = hour * 24
            out = list()
            for a, b in [(day, '天'), (hour, '小时'), (minute, '分')]:
                cost = 0
                while t > a:
                    cost += 1
                    t -= a
                if cost > 0:
                    out.append(f'{cost}{b}')
            out.append('{}秒'.format(math.ceil(t / second)))
            return ''.join(out)

        status = self.session.get(self.host + '/jobs/' + job_id).json()
        name = status['name'].split(':', maxsplit=1)[0]
        if '.' in name:
            name, pt = name.split('.', maxsplit=1)
        else:
            pt = None
        job_status = JobStatus(name, job_id, status['state'], full_name=status['name'],
                               start_time=p_time(status["start-time"]), end_time=p_time(status["end-time"]),
                               duration=p_duration(status['duration']), pt=pt)
        return job_status

    @property
    def job_status(self) -> List[JobStatus]:
        jobs_url = self.host + '/jobs'
        js = self.session.get(jobs_url).json()
        all_jobs = list()
        for x in js['jobs']:
            j_id = x['id']
            if self.cache.get(j_id) is None:
                status = self._get_job_status(j_id)
                self.cache.set(j_id, status)
            else:
                status = self.cache.get(j_id)

            all_jobs.append(status)
        return all_jobs

    @property
    def live_job_names(self) -> Set[str]:
        data = set()
        for job in self.job_status:
            if job.status == 'RUNNING':
                data.add(job.name)
        return data

    def handle_restart(self, transform: Transform, **kwargs) -> str:
        msgs = []
        msgs.append(self.handle_stop(transform))
        msgs.append(self.start_flink_job(transform, **kwargs))
        return SUCCESS_HEADER + '\n'.join(msgs)

    def handle_stop(self, transform: Transform, **kwargs) -> str:
        msgs = []
        header = get_job_short_name(transform)
        kill_jobs = []
        job_status = self.job_status
        pt = kwargs['pt'] if 'pt' in kwargs else None
        kill_all_pt = True if 'kill_all_pt' in kwargs else False
        for job in job_status:
            if job.status == self.RUN_STATUS and job.name == header and (job.pt == pt or kill_all_pt):
                logger.debug('add a {} to kill '.format(job.name))
                kill_jobs.append(job.job_id)

        msgs.append('kill {} jobs: {}'.format(len(kill_jobs), ', '.join(str(x) for x in kill_jobs)))

        self.stop_flink_jobs(kill_jobs)
        return SUCCESS_HEADER + '\n'.join(msgs)

    def handle_start(self, transform: Transform, **kwargs) -> str:
        return self.start_flink_job(transform, **kwargs)

    def handle_cancel(self, jid: str, **kwargs):
        self.stop_flink_jobs([jid])
        return 'kill {} '.format(jid)

    def handle_status(self, transform: Transform, **kwargs) -> str:
        header = get_job_short_name(transform)
        job_status = self.job_status
        pt = kwargs['pt'] if 'pt' in kwargs else None
        last_run_job_id = kwargs['last_run_job_id'].split('_') if 'last_run_job_id' in kwargs else []
        if last_run_job_id:

            all_job = list(filter(lambda x: x.job_id in last_run_job_id, job_status))
            if len(all_job) == 0:
                logger.error('{} last run job id not found in flink job {}'.format(header, str(last_run_job_id)))
                return self.FAIL_STATUS
            if any(map(lambda x: x.status not in (self.RUN_STATUS, self.FINISHED_STATUS), all_job)):
                logger.debug('{} find  one of the job failed {}'.format(header, str(last_run_job_id)))
                return self.FAIL_STATUS
            if any(map(lambda x: x.status == self.RUN_STATUS, all_job)):
                return '_'.join(last_run_job_id + [self.RUN_STATUS])
            if all(map(lambda x: x.status == self.FINISHED_STATUS, all_job)):
                return self.FINISHED_STATUS
            logger.error('{} some thing wrong with last run job id {}'.format(header, str(last_run_job_id)))
            return self.FAIL_STATUS

        statuses = Counter()
        last_end_time = None
        for job in job_status:
            if job.name == header and job.pt == pt:
                statuses[job.status] += 1
                if job.status == self.RUN_STATUS:
                    last_run_job_id.append(job.job_id)

                if job.status == self.FINISHED_STATUS:
                    end_time = datetime.strptime(job.end_time, '%Y-%m-%d %H:%M:%S')
                    if last_end_time is None or end_time > last_end_time:
                        last_end_time = end_time

        if statuses[self.RUN_STATUS] > 0:
            return '_'.join(last_run_job_id + [self.RUN_STATUS])
        if statuses[self.FINISHED_STATUS] > 0:
            if (datetime.now() - last_end_time).seconds < 20:
                logger.debug('{} job finished too fast'.format(header))
                return self.FINISHED_STATUS
            else:
                logger.error('{} job not found latest finished job'.format(header))
                return self.FAIL_STATUS

        logger.error('{} job create job success but not found any job info in flink'.format(header))
        return self.FAIL_STATUS

    def stop_flink_jobs(self, job_ids: List):
        for j_id in job_ids:
            logger.debug('begin stop flink job {}'.format(j_id))
            logger.debug(self.host)
            res = self.session.patch(self.host + '/jobs/' + j_id + '?mode=cancel')
            self.cache.remove(j_id)
            print(res.text)

    @classmethod
    def start_flink_job(cls, transform: Transform, **kwargs) -> str:
        is_ok, txt = run_transform(transform, **kwargs)
        return '{} JOB {}\n{}'.format(SUCCESS_HEADER if is_ok else FAIL_HEADER, transform.name, '' if is_ok else txt)

    job_pattern = re.compile(r'\d+_')

    @classmethod
    def is_real_job(cls, name) -> bool:
        return cls.job_pattern.search(name) is not None


JobControlHandle = JobControl(FSQLFLY_FINK_HOST)


def _handle_job(mode: str, pk: str, json_body: dict, session: Session) -> DBRes:
    handle_name = 'handle_' + mode

    if mode in JobControlHandle and handle_name in JobControlHandle:
        if pk.isdigit():
            transform = DBDao.get_transform(pk, session=session)
            if transform is None:
                return DBRes.api_error(msg='job id {} not found!!!'.format(pk))
        else:
            transform = pk
        data = json_body
        run_res = getattr(JobControlHandle, handle_name)(transform, **data)
        return DBRes(code=500 if run_res.startswith(FAIL_HEADER) else 200, msg=run_res)
    else:
        return DBRes.api_error(msg=' {} not support!!!'.format(mode))


def handle_job(mode: str, pk: str, json_body: dict) -> DBRes:
    session = DBSession.get_session()
    try:
        return _handle_job(mode, pk, json_body, session)
    finally:
        session.close()



import tornado.web
from terminado import TermSocket
from fsqlfly.common import safe_authenticated
from fsqlfly import settings
from fsqlfly.base_handle import BaseHandler
from fsqlfly.workflow import run_debug_transform
from fsqlfly.common import DBRes
from fsqlfly.utils.job_manage import handle_job
from fsqlfly.db_helper import DBDao


class TerminalHandler(BaseHandler):
    @safe_authenticated
    def get(self):
        tm = self.terminal_manager
        terms = [{'name': name, 'id': name} for name in tm.terminals]
        self.write_res(DBRes(data=terms))


class TransformControlHandler(BaseHandler):
    @safe_authenticated
    def post(self, mode: str, pk: str):
        if mode == 'debug':
            term = run_debug_transform(self.json_body, self.terminal_manager)
            self.write_res(DBRes({"url": '/terminal/{}'.format(term)}))
        else:
            if not pk.isdigit():
                pk = str(DBDao.name2pk(model='transform', name=pk))

            return self.write_res(handle_job(mode, pk, self.json_body))


class TerminalStopHandler(BaseHandler):
    @safe_authenticated
    async def post(self, name: str):
        tm = self.terminal_manager
        if name in tm.terminals:
            await tm.terminate(name, force=True)
            self.write_res(DBRes())
        else:
            raise tornado.web.HTTPError(404, "Terminal not found: %r" % name)


class MyTermSocket(TermSocket):
    def open(self, url_component=None):
        super(MyTermSocket, self).open(url_component)
        terminal = self.term_manager.get_terminal(url_component)
        open_name = settings.TERMINAL_OPEN_NAME
        if hasattr(terminal, open_name) and getattr(terminal, open_name):
            setattr(terminal, open_name, False)
            self.term_manager.start_reading(terminal)


default_handlers = [
    (r'/api/terminal', TerminalHandler),
    (r"/_websocket/(\w+)", MyTermSocket, {'term_manager': settings.TERMINAL_MANAGER}),
    (r'/api/terminal/stop/(?P<name>\d+)', TerminalStopHandler),
    (r'/api/transform/(\w+)/([0-9a-zA-Z_]+)', TransformControlHandler),
]



import unittest
from fsqlfly.db_helper import *
from fsqlfly.tests.base_test import FSQLFlyTestCase


class MyTestCase(FSQLFlyTestCase):
    def test_topic_reload(self):
        kafak_connector = """
properties:
- key: bootstrap.servers
  value: localhost:9092
- key: group.id
  value: {{ resource_name.database }}__{{ resource_name.name }}_group
property-version: 1
startup-mode: {{ startup_mode | default('latest-offset') }}
topic: {{ resource_name.database }}__{{ resource_name.name }}__{{ version.name }}
version: universal
        """
        real_topic = 'abcdefg'
        config = f'''
[kafka]
topic: {real_topic}
        '''
        connection = Connection(name='a', url='#', type='kafka', connector=kafak_connector)
        r_name = ResourceName(name='b', full_name='a.b', database='dd', connection=connection,
                              config=config)
        r2_name = ResourceName(name='b2', full_name='a.b2', database='dd', connection=connection)
        t_name = ResourceTemplate(name='c', resource_name=r_name, type='both', full_name='a.b.c', connection=connection)
        t2_name = ResourceTemplate(name='c2', resource_name=r2_name, type='both', full_name='a.b.c2',
                                   connection=connection)
        v_name = ResourceVersion(name='d', template=t_name, full_name='a.b.c.d', connection=connection,
                                 resource_name=r_name)
        v2_name = ResourceVersion(name='d', template=t2_name, full_name='a.b.c.d2', connection=connection,
                                  resource_name=r2_name)
        session = DBSession.get_session()

        session.add_all([connection, r_name, r2_name, t_name, t2_name, v_name, v2_name])
        session.commit()
        # cache = session.query(ResourceVersion).filter(ResourceVersion.id==v_name.id).one().generate_version_cache()
        cache = v_name.generate_version_cache()
        # cache2 = session.query(ResourceVersion).filter(ResourceVersion.id==v2_name.id).one().generate_version_cache()
        cache2 = v2_name.generate_version_cache()
        self.assertEqual(cache2['connector']['topic'], 'dd__b2__d')
        self.assertEqual(cache['connector']['topic'], real_topic)


if __name__ == '__main__':
    unittest.main()



import unittest
from fsqlfly.contrib.mail import MailHelper

class MyTestCase(unittest.TestCase):
    def test_send(self):
        MailHelper.send('test', 'hello')


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch
from fsqlfly.db_helper import *
from fsqlfly.tests.base_test import FSQLFlyTestCase


class MyTestCase(FSQLFlyTestCase):

    def test_positive_delete(self):
        namespace = Namespace(name='iii')
        self.session.add(namespace)
        self.session.commit()

        t = Transform(name='test', sql='select 1;', namespace=namespace)
        self.session.add(t)
        self.session.commit()
        self.session.delete(namespace)
        self.session.commit()
        self.assertEqual(self.session.query(Transform).count(), 0)

    def get_create_object(self):
        connection = Connection(name='a', url='#', type='hive', connector='text')
        schema = SchemaEvent(name='test', connection=connection, version=1)
        schema2 = SchemaEvent(name='test2', connection=connection, version=2)
        r_name = ResourceName(name='b', full_name='a.b', connection=connection, schema_version=schema)
        t_name = ResourceTemplate(name='c', resource_name=r_name, type='both', full_name='a.b.c', connection=connection,
                                  schema_version=schema)
        v_name = ResourceVersion(name='d', template=t_name, full_name='a.b.c.d', connection=connection,
                                 resource_name=r_name, schema_version=schema)
        return connection, schema, schema2, r_name, t_name, v_name

    def test_positive_delete_connection(self):
        connection, schema, schema2, r_name, t_name, v_name = self.get_create_object()

        self.session.add_all([connection, schema, schema2, r_name, t_name, v_name])
        self.session.commit()
        self.session.delete(connection)
        self.session.commit()
        self.assertEqual(self.session.query(ResourceName).count(), 0)
        self.assertEqual(self.session.query(Connection).count(), 0)

        self.assertEqual(self.session.query(ResourceVersion).count(), 0)
        self.assertEqual(self.session.query(ResourceTemplate).count(), 0)
        self.assertEqual(self.session.query(SchemaEvent).count(), 0)

    def test_positive_delete_connection_by_db_helper(self):
        connection, schema, schema2, r_name, t_name, v_name = self.get_create_object()

        self.session.add_all([connection, schema, schema2, r_name, t_name, v_name])
        self.session.commit()
        self.assertEqual(self.session.query(Connection).count(), 1)
        DBSession.init_engine(self.engine)
        with patch.object(settings, 'FSQLFLY_SAVE_MODE_DISABLE', True):
            res = DBDao.delete('connection', pk=connection.id)
        self.assertEqual(res.success, True)
        self.session.close()
        self.session = self.get_session()
        self.assertEqual(self.session.query(Connection).count(), 0)
        self.assertEqual(self.session.query(ResourceName).count(), 0)
        self.assertEqual(self.session.query(ResourceVersion).count(), 0)
        self.assertEqual(self.session.query(ResourceTemplate).count(), 0)
        self.assertEqual(self.session.query(SchemaEvent).count(), 0)

    def test_positive_delete_other(self):
        connection, schema, schema2, r_name, t_name, v_name = self.get_create_object()

        self.session.add_all([connection, schema, schema2, r_name, t_name, v_name])
        self.session.commit()
        self.session.delete(schema)

        self.session.commit()
        self.assertEqual(self.session.query(Connection).count(), 1)
        self.assertEqual(self.session.query(ResourceName).count(), 0)
        self.assertEqual(self.session.query(Connection).count(), 1)

        self.assertEqual(self.session.query(ResourceVersion).count(), 0)
        self.assertEqual(self.session.query(ResourceTemplate).count(), 0)
        self.assertEqual(self.session.query(SchemaEvent).count(), 1)

    def test_get_connection_and_resource_name_config(self):
        connection_config = """
[jdbc]
insert_primary_key = false

        """
        resource_name_config = """
[jdbc]
insert_primary_key = true
        """
        connection = Connection(name='a', url='#', type='hive', connector='text', config=connection_config)
        schema = SchemaEvent(name='test', connection=connection)
        r_name = ResourceName(name='b', full_name='a.b', connection=connection, schema_version=schema,
                              config=resource_name_config)
        self.assertTrue(not r_name.get_config('add_read_partition_key', 'jdbc', bool))
        self.assertTrue(not r_name.get_config('add_read_partition_key', 'jdbc', bool))
        self.assertEqual(connection.get_config('read_partition_num', 'jdbc', int), 50)
        self.assertTrue(r_name.get_config('example11') is None)

        self.assertTrue(r_name.get_config('insert_primary_key', 'jdbc', bool))
        self.assertTrue(not connection.get_config('insert_primary_key', 'jdbc', bool))


if __name__ == '__main__':
    unittest.main()



import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from datetime import datetime, date


class MailSender:
    def __init__(self, mail_host: str, mail_user: str, mail_password: str, mail_receivers: str):
        self.mail_host = mail_host
        self.mail_user = mail_user
        self.mail_password = mail_password
        self.mail_receivers = mail_receivers.split(',')

    def send_email(self, title: str, content: str):
        today = date.today()
        title = f'{today}{title}'

        message = MIMEMultipart()
        message['From'] = "{}".format(self.mail_user)
        message['To'] = ",".join(self.mail_receivers)
        message['Subject'] = title
        txt = MIMEText(content, 'plain', 'utf-8')
        message.attach(txt)

        try:
            smtp_obj = smtplib.SMTP_SSL(self.mail_host, 465)
            smtp_obj.login(self.mail_user, self.mail_password)
            smtp_obj.sendmail(self.mail_user, self.mail_receivers, message.as_string())
            return "{} mail has been send successfully.".format(str(datetime.now())[:19])
        except smtplib.SMTPException as e:
            return str(e)


class MailHelper:
    _sender: MailSender = None

    @classmethod
    def init(cls):
        from fsqlfly.settings import (FSQLFLY_MAIL_HOST, FSQLFLY_MAIL_USER, FSQLFLY_MAIL_PASSWORD,
                                      FSQLFLY_MAIL_RECEIVERS)
        cls._sender = MailSender(FSQLFLY_MAIL_HOST,
                                 FSQLFLY_MAIL_USER,
                                 FSQLFLY_MAIL_PASSWORD,
                                 FSQLFLY_MAIL_RECEIVERS)

    @classmethod
    def send(cls, title: str, content: str):
        if cls._sender is None:
            cls.init()

        cls._sender.send_email(title, content)



# -*- coding: utf-8 -*-
from datetime import datetime, timedelta


def ds_add(ds, days):
    """
    Add or subtract days from a YYYY-MM-DD

    :param ds: anchor date in ``YYYY-MM-DD`` format to add to
    :type ds: str
    :param days: number of days to add to the ds, you can use negative values
    :type days: int

    >>> ds_add('2015-01-01', 5)
    '2015-01-06'
    >>> ds_add('2015-01-06', -5)
    '2015-01-01'
    """

    ds = datetime.strptime(ds, '%Y-%m-%d')
    if days:
        ds = ds + timedelta(days)
    return ds.isoformat()[:10]


def ds_format(ds, input_format, output_format):
    """
    Takes an input string and outputs another string
    as specified in the output format

    :param ds: input string which contains a date
    :type ds: str
    :param input_format: input string format. E.g. %Y-%m-%d
    :type input_format: str
    :param output_format: output string format  E.g. %Y-%m-%d
    :type output_format: str

    >>> ds_format('2015-01-01', "%Y-%m-%d", "%m-%d-%y")
    '01-01-15'
    >>> ds_format('1/5/2015', "%m/%d/%Y",  "%Y-%m-%d")
    '2015-01-05'
    """
    return datetime.strptime(ds, input_format).strftime(output_format)



import sqlalchemy as sa
from jinja2 import Template
from datetime import datetime
from configparser import ConfigParser
from typing import Tuple, TypeVar
from sqlalchemy import Column, String, ForeignKey, Integer, DateTime, Boolean, Text, UniqueConstraint
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declarative_base
from fsqlfly.common import *

from fsqlfly.utils.strings import load_yaml
from fsqlfly.utils.template import generate_template_context
from fsqlfly.utils.db_execute import execute
from sqlalchemy_utils import ChoiceType, Choice
from logzero import logger

_Base = declarative_base()

CONNECTION_TYPE = ChoiceType([(k, k) for k in SUPPORT_MANAGER], impl=String(16))
TABLE_TYPE = ChoiceType([(k, k) for k in SUPPORT_TABLE_TYPE], impl=String(8))
CONNECTOR_CHOICE_TYPE = ChoiceType([(k, k) for k in CONNECTOR_TYPE], impl=String(8))
CONFIG_T = TypeVar('CONFIG_T', int, str, bool, float)


def _b(x: str):
    return backref(x, cascade="delete, delete-orphan")


class SaveDict(dict):
    @property
    def id(self):
        return self.get('id')


class Base(_Base):
    __abstract__ = True

    id = Column(Integer, primary_key=True, autoincrement=True)
    created_at = Column(DateTime, server_default=sa.func.now())
    updated_at = Column(DateTime, server_default=sa.func.now(), server_onupdate=sa.func.now())
    is_locked = Column(Boolean, default=False)

    def as_dict(self) -> SaveDict:
        def _convert(v: Any) -> Any:
            if isinstance(v, Choice):
                return v.code
            return v

        return SaveDict({column.name: _convert(getattr(self, column.name)) for column in self.__table__.columns})

    @classmethod
    def get_default_config_parser(cls) -> ConfigParser:
        default_config_parser = ConfigParser()

        default_config_parser.read_string(DEFAULT_CONFIG)
        return default_config_parser

    def get_config_parser(self) -> ConfigParser:
        return self.get_default_config_parser()

    def get_config(self, name: str, section: str, typ: Optional[Type[CONFIG_T]] = None) -> CONFIG_T:
        cache_name = '__config_parser__'
        if not hasattr(self, cache_name):
            setattr(self, cache_name, self.get_config_parser())
        parser = getattr(self, cache_name)
        secret = parser[section]
        if typ:
            if typ is int:
                return secret.getint(name)
            elif typ is float:
                return secret.getint(name)
            elif typ is bool:
                return secret.getboolean(name)
        return secret.get(name)


class Connection(Base):
    __tablename__ = 'connection'
    name = Column(String(32), nullable=False, unique=True)
    update_interval = Column(Integer, default=0)
    url = Column(Text, nullable=False)
    type = Column(CONNECTION_TYPE, nullable=False)
    info = Column(Text)
    connector = Column(Text, nullable=False)
    config = Column(Text)
    include = Column(String(2048))
    exclude = Column(String(2048))
    is_active = Column(Boolean, default=False)

    def get_config(self, name: str, section: Optional[str] = None,
                   typ: Optional[Union[Type[int], Type[float], Type[float], Type[str]]] = None) -> Any:
        return super(Connection, self).get_config(name, section=section if section else self.type, typ=typ)

    def get_config_parser(self) -> ConfigParser:
        parser = super(Connection, self).get_config_parser()
        if self.config and self.config.strip():
            parser.read_string(self.config)

        return parser

    def get_connection_connector(self) -> dict:
        config = self.get_config_parser()
        context = generate_template_context(execution_date=datetime.now(), connection=self, **config[self.type])
        output = Template(self.connector).render(**context)
        return load_yaml(output)


class SchemaEvent(Base):
    __tablename__ = 'schema_event'
    __table_args__ = (
        UniqueConstraint('connection_id', 'name', 'database', 'version'),
    )
    name = Column(String(128), nullable=False)
    info = Column(Text)
    database = Column(String(64), nullable=True)
    connection_id = Column(Integer, ForeignKey('connection.id'), nullable=False)
    connection = relationship('Connection', backref=_b('schemas'))
    father_id = Column(Integer, ForeignKey('schema_event.id'), nullable=True)
    father = relationship('SchemaEvent', backref=_b('children'), remote_side='SchemaEvent.id')
    version = Column(Integer, nullable=False, default=0)
    comment = Column(Text)
    primary_key = Column(String(64))
    fields = Column(Text)
    partitionable = Column(Boolean, default=False)


class Connector(Base):
    __tablename__ = 'connector'
    name = Column(String(128), nullable=False, unique=True)
    info = Column(Text)
    type = Column(CONNECTOR_CHOICE_TYPE, nullable=False)
    source_id = Column(Integer, ForeignKey('connection.id', ondelete='NO ACTION'), nullable=True)
    target_id = Column(Integer, ForeignKey('connection.id', ondelete='NO ACTION'), nullable=True)
    source = relationship(Connection, foreign_keys=source_id, passive_deletes=False)
    target = relationship(Connection, foreign_keys=target_id, passive_deletes=False)
    config = Column(Text)
    generate_sql = Column(Text)
    cache = Column(Text)

    def get_config_parser(self) -> ConfigParser:
        parser = self.get_default_config_parser()
        if self.config and self.config.strip():
            parser.read_string(self.config)

        return parser

    @property
    def connector_mode(self) -> CanalMode:
        config_mode = self.get_config('mode', typ=str)

        return CanalMode(config_mode)

    def get_config(self, name: str, section: Optional[str] = None, typ: Optional[Type[CONFIG_T]] = None) -> CONFIG_T:
        return super(Connector, self).get_config(name, section=section if section else self.type.code, typ=typ)

    @property
    def process_time_field(self) -> Optional[SchemaField]:
        if self.get_config('process_time_enable', typ=bool):
            return SchemaField(name=self.get_config('process_time_name', typ=str),
                               type=BlinkSQLType.TIMESTAMP, proctime=True)

    @property
    def rowtime_enable(self) -> bool:
        return self.get_config('rowtime_enable', typ=bool)

    @property
    def rowtime_field(self) -> Optional[SchemaField]:
        if self.get_config('rowtime_enable', typ=bool):
            row_field = {
                "timestamps": {"type": "from-field", "from": self.get_config('rowtime_from', typ=str)},
                "watermarks": {"type": "periodic-bounded", "delay": self.get_config("rowtime_watermarks", typ=int)}
            }
            return SchemaField(name=self.get_config('rowtime_name', typ=str),
                               type=BlinkSQLType.TIMESTAMP, rowtime=row_field)

    @property
    def binlog_type_name_field(self) -> SchemaField:
        return SchemaField(name=self.get_config('binlog_type_name', typ=str),
                           type=BlinkSQLType.INTEGER)

    @property
    def db_execute_time_field(self) -> Optional[SchemaField]:
        if not self.get_config('rowtime_enable', typ=bool):
            return SchemaField(name=self.get_config('rowtime_from', typ=str),
                               type=BlinkSQLType.TIMESTAMP)

    @property
    def binlog_type_name(self) -> str:
        return self.get_config('binlog_type_name', typ=str)

    @property
    def before_column_suffix(self) -> str:
        return self.get_config('before_column_suffix', typ=str)

    @property
    def after_column_suffix(self) -> str:
        return self.get_config('after_column_suffix', typ=str)

    @property
    def update_suffix(self) -> str:
        return self.get_config('update_suffix', typ=str)

    def check_system_type(self):
        assert self.type == 'system', 'only system has need tables '

    @property
    def need_tables(self) -> NameFilter:
        self.check_system_type()
        return NameFilter(self.get_config('source_include'), self.get_config('source_exclude'))

    def get_transform_name_format(self, resource_name: 'ResourceName', **kwargs) -> str:
        self.check_system_type()
        source_type, target_type = self.source.type.code, self.target.type.code
        return Template(self.get_config('transform_name_format')).render(
            generate_template_context(source_type=source_type, target_type=target_type,
                                      resource_name=resource_name, connector=self, **kwargs))

    def get_transform_target_full_name(self, **kwargs) -> Tuple[str, str]:
        database = Template(self.get_config('target_database_format')).render(**kwargs)
        table = Template(self.get_config('target_table_format')).render(**kwargs)
        return database, table

    @property
    def use_partition(self) -> bool:
        self.check_system_type()
        return self.get_config('use_partition', typ=bool)

    @property
    def system_overwrite(self) -> bool:
        self.check_system_type()
        return self.get_config('overwrite', typ=bool)

    @property
    def system_execution_parallelism(self) -> int:
        self.check_system_type()
        return self.get_config('execution_parallelism', typ=int)

    @property
    def partition_key_value(self) -> Tuple[str, str]:
        self.check_system_type()
        return self.get_config('partition_name'), self.get_config('partition_value')


class ResourceName(Base):
    __tablename__ = 'resource_name'
    __table_args__ = (
        UniqueConstraint('connection_id', 'name', 'database'),
    )
    name = Column(String(128), nullable=False)
    info = Column(Text)
    database = Column(String(64), nullable=True)
    connection_id = Column(Integer, ForeignKey('connection.id'), nullable=False)
    connection = relationship(Connection, backref=_b('resource_names'))
    schema_version_id = Column(Integer, ForeignKey('schema_event.id'), nullable=True)
    schema_version = relationship(SchemaEvent, backref=_b('resource_names'), foreign_keys=schema_version_id)
    latest_schema_id = Column(Integer, ForeignKey('schema_event.id'), nullable=True)
    config = Column(Text)
    is_latest = Column(Boolean, default=True)
    is_active = Column(Boolean, default=True)
    full_name = Column(String(512), nullable=False, unique=True)

    def get_include(self):
        return self.database + '.' + self.name if self.database else self.name

    def get_config_parser(self) -> ConfigParser:
        parser = self.connection.get_config_parser()
        if self.config and self.config.strip():
            parser.read_string(self.config)

        return parser

    def get_config(self, name: str, section: Optional[str] = None,
                   typ: Optional[Union[Type[CONFIG_T]]] = None) -> CONFIG_T:
        return super(ResourceName, self).get_config(name, section=section if section else self.connection.type, typ=typ)

    @property
    def db_name(self):
        return self.database + '.' + self.name if self.database else self.name


class ResourceTemplate(Base):
    __tablename__ = 'resource_template'
    __table_args__ = (
        UniqueConstraint('resource_name_id', 'name'),
    )
    name = Column(String(128), nullable=False)
    type = Column(TABLE_TYPE, nullable=False)
    info = Column(Text)
    is_system = Column(Boolean, default=False)
    is_default = Column(Boolean, default=False)
    full_name = Column(String(512), nullable=False, unique=True)
    connection_id = Column(Integer, ForeignKey('connection.id'), nullable=False)
    connection = relationship('Connection', backref=_b('templates'))
    resource_name_id = Column(Integer, ForeignKey('resource_name.id'), nullable=False)
    resource_name = relationship(ResourceName, backref=_b('templates'))
    schema_version = relationship(SchemaEvent, backref=_b('templates'))
    schema_version_id = Column(Integer, ForeignKey('schema_event.id'), nullable=True)
    config = Column(Text)


class ResourceVersion(Base):
    __tablename__ = 'resource_version'
    __table_args__ = (
        UniqueConstraint('template_id', 'version'),
    )
    name = Column(String(128), nullable=False, default='latest')
    info = Column(Text)
    full_name = Column(String(512), nullable=False, unique=True)
    is_system = Column(Boolean, default=False)
    is_default = Column(Boolean, default=False)
    is_latest = Column(Boolean, default=False)
    version = Column(Integer, nullable=False, default=0)
    connection_id = Column(Integer, ForeignKey('connection.id'), nullable=False)
    connection = relationship(Connection, backref=_b('versions'))
    resource_name_id = Column(Integer, ForeignKey('resource_name.id'), nullable=False)
    resource_name = relationship(ResourceName, backref=_b('versions'))
    template_id = Column(Integer, ForeignKey('resource_template.id'), nullable=False)
    template = relationship(ResourceTemplate, backref=_b('versions'))
    schema_version = relationship(SchemaEvent, backref=_b('versions'))
    schema_version_id = Column(Integer, ForeignKey('schema_event.id'))
    config = Column(Text)
    cache = Column(Text)

    def get_connection_connector(self) -> dict:
        config = self.resource_name.get_config_parser()
        context = generate_template_context(execution_date=datetime.now(), connection=self.connection,
                                            schema=self.schema_version,
                                            tempalte=self.template, resource_name=self.resource_name,
                                            version=self, **config[self.connection.type])
        output = Template(self.connection.connector).render(**context)
        return load_yaml(output)

    @classmethod
    def latest_name(cls):
        return 'latest'

    @property
    def db_full_name(self):
        return self.full_name.replace('.', '__')

    def get_db_bound(self, key: str) -> Tuple[int, int]:
        resource_name = self.resource_name
        connection = self.connection

        def _get_int(name: str):
            return resource_name.get_config(name, typ=int)

        if resource_name.get_config('auto_partition_bound', typ=bool):
            sql = f'select min(`{key}`), max(`{key}`) from `{resource_name.database}`.`{resource_name.name}`'
            data = execute(sql, connection.url)
            l, u = data[0]
        else:
            l, u = _get_int('read_partition_lower_bound'), _get_int('read_partition_upper_bound')
        return l, u

    def generate_version_cache(self) -> dict:
        template = self.template
        version = self
        resource_name = self.resource_name
        connection = self.connection
        connection_type = connection.type.code
        template_type = template.type.code
        schema = self.schema_version if self.schema_version else SchemaContent(name=resource_name.name,
                                                                               type=connection.type.code,
                                                                               database=resource_name.database)
        base = load_yaml(template.config) if template.config else dict()
        version_config = load_yaml(version.config) if version.config else dict()
        base.update(**version_config)
        config = VersionConfig(**base)
        v_info = "Version {}:{}".format(version.id, version.name)
        if template.type == 'view':
            assert config.query is not None, '{} View Must Need a Query'.format(v_info)
            res = dict(query=config.query)
        elif template.type == 'temporal-table':
            assert config.history_table is not None, '{} View Must Need a history_table'.format(v_info)
            assert config.primary_key is not None, '{} View Must Need a primary_key'.format(v_info)
            assert config.time_attribute is not None, '{} View Must Need a time_attribute'.format(v_info)
            res = {
                "history-table": config.history_table,
                "primary-key": config.primary_key,
                "time-attribute": config.time_attribute
            }
        elif template.type in ('sink', 'source', 'both'):
            res = dict()
            if config.update_mode:
                res['update-mode'] = config.update_mode
            if config.format:
                res['format'] = config.format
            elif connection.type.code not in ['hbase', 'jdbc']:
                res['format'] = {"type": 'json', "derive-schema": True}

            connector = version.get_connection_connector()

            if connector and template.type in ('source', 'both') and schema.primary_key and connection_type == 'jdbc':
                if resource_name.get_config('add_read_partition_key', typ=bool) and 'read' not in connector:
                    key = resource_name.get_config('read_partition_key')
                    if key is None and schema.primary_key:
                        key = schema.primary_key
                    if key:
                        low, upper = self.get_db_bound(key)
                        connector['read'] = {
                            "partition": {
                                "column": key,
                                "num": resource_name.get_config('read_partition_num', typ=int),
                                "lower-bound": low,
                                "upper-bound": upper,
                            },
                            "fetch-size": resource_name.get_config('read_partition_fetch_size', typ=int)
                        }
            if connector and connection_type == 'kafka':
                if resource_name.get_config('topic'):
                    connector['topic'] = resource_name.get_config('topic')

            if connector:
                connector['type'] = connection.type.code

            res['connector'] = connector if connector else None

            fields = [SchemaField(**x) for x in load_yaml(schema.fields)] if schema.fields else []
            need_fields = [x for x in fields if x.name in NameFilter(config.include, config.exclude)]
            field_names = [x.name for x in need_fields]
            schemas = []

            if config.schema:
                for x in config.schema:
                    assert x['name'] not in field_names, "{} contain in origin field"
                schemas.extend(config.schema)

            for x in need_fields:
                if x.autoincrement and connection_type == 'jdbc' and template_type == 'sink' and resource_name.get_config('insert_primary_key', typ=bool):
                    continue
                schemas.append(self.field2schema(x))

            res['schema'] = schemas

        else:
            raise NotImplementedError("Not Support {}".format(template.type))

        return res

    @classmethod
    def field2schema(cls, field: SchemaField) -> dict:
        base = {
            "name": field.name,
            "data-type": field.type
        }
        if field.proctime:
            base['proctime'] = field.proctime
        if field.rowtime:
            base['rowtime'] = field.rowtime
        return base


class Namespace(Base):
    __tablename__ = 'namespace'
    name = Column(String(256), unique=True)
    info = Column(Text)
    avatar = Column(Text)
    is_daemon = Column(Boolean, default=False)


class FileResource(Base):
    __tablename__ = 'file_resource'
    name = Column(String(512), unique=True, nullable=False)
    info = Column(Text)
    real_path = Column(Text)
    is_system = Column(Boolean, default=False)


class Functions(Base):
    __tablename__ = 'functions'
    name = Column(String(256), unique=True, nullable=False)
    class_name = Column(String(512), nullable=False)
    constructor_config = Column(Text)
    resource_id = Column(Integer, ForeignKey('file_resource.id'), nullable=False)
    resource = relationship(FileResource, backref=_b('functions'))
    is_active = Column(Boolean, default=True)

    class Meta:
        table_name = "functions"


class Transform(Base):
    __tablename__ = 'transform'
    name = Column(String(256), unique=True, nullable=False)
    info = Column(Text)
    sql = Column(Text, nullable=True)
    require = Column(Text)
    yaml = Column(Text)
    namespace_id = Column(Integer, ForeignKey('namespace.id'), nullable=True)
    namespace = relationship(Namespace, backref=_b('transforms'))
    connector_id = Column(Integer, ForeignKey('connector.id'), nullable=True)
    connector = relationship(Connector, backref=_b('transforms'))
    is_daemon = Column(Boolean, default=False)

    class Meta:
        table_name = "transform"


class TransformSavepoint(Base):
    __tablename__ = 'transform_savepoint'
    name = Column(String(1024), nullable=False)
    path = Column(Text, nullable=False)
    info = Column(Text)
    snapshot = Column(Text)
    transform_id = Column(Integer, ForeignKey('transform.id'))
    transform = relationship('Transform', backref=_b('savepoint'))


def delete_all_tables(engine, force: bool = False):
    if not force:
        word = input('Are you delete all tables (Y/n)')
        if word.strip().upper() != 'Y':
            return
    logger.info("begin delete all tables")
    Base.metadata.drop_all(engine)


def create_all_tables(engine):
    Base.metadata.create_all(engine)


__all__ = ['create_all_tables', 'delete_all_tables', 'Base', 'Connection',
           'SchemaEvent', 'Connector', 'ResourceName', 'ResourceVersion', 'ResourceTemplate',
           'Namespace', 'FileResource', 'Transform', 'Functions', 'TransformSavepoint', 'SaveDict']
